#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import json as json_lib
from unittest.mock import MagicMock

import pytest
import pytest_twisted
from twisted.web import server
from twisted.web.http import Request

import deluge.common
import deluge.ui.web.auth
import deluge.ui.web.json_api
from deluge.error import DelugeError
from deluge.ui.web.auth import Auth
from deluge.ui.web.json_api import JSON, JSONException

from . import common
from .common_web import WebServerMockBase

common.disable_new_release_check()


@pytest.mark.usefixtures('daemon', 'client', 'component')
class TestJSON:
    @pytest_twisted.ensureDeferred
    async def test_get_remote_methods(self):
        json = JSON()
        methods = await json.get_remote_methods()
        assert type(methods) == tuple
        assert len(methods) > 0

    def test_render_fail_disconnected(self):
        json = JSON()
        request = MagicMock()
        request.method = b'POST'
        request._disconnected = True
        # When disconnected, returns empty string
        assert json.render(request) == ''

    def test_render_fail(self):
        json = JSON()
        request = MagicMock()
        request.method = b'POST'

        def write(response_str):
            request.write_was_called = True
            response = json_lib.loads(response_str.decode())
            assert response['result'] is None
            assert response['id'] is None
            assert response['error']['message'] == 'JSONException: JSON not decodable'
            assert response['error']['code'] == 5

        request.write = write
        request.write_was_called = False
        request._disconnected = False
        request.getHeader.return_value = b'application/json'
        assert json.render(request) == server.NOT_DONE_YET
        assert request.write_was_called

    def test_handle_request_invalid_method(self):
        json = JSON()
        request = MagicMock()
        json_data = {'method': 'no-existing-module.test', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        request_id, result, error = json._handle_request(request)
        assert error == {'message': 'Unknown method', 'code': 2}

    def test_handle_request_invalid_json_request(self):
        json = JSON()
        request = MagicMock()
        json_data = {'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        with pytest.raises(JSONException):
            json._handle_request(request)
        json_data = {'method': 'some.method', 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        with pytest.raises(JSONException):
            json._handle_request(request)
        json_data = {'method': 'some.method', 'id': 0}
        request.json = json_lib.dumps(json_data).encode()
        with pytest.raises(JSONException):
            json._handle_request(request)

    def test_on_json_request_invalid_content_type(self):
        """Test for exception with content type not application/json"""
        json = JSON()
        request = MagicMock()
        request.getHeader.return_value = b'text/plain'
        json_data = {'method': 'some.method', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        with pytest.raises(JSONException):
            json._on_json_request(request)

    def test_on_json_request_valid_content_type(self):
        """Ensure content-type application/json is accepted"""
        json = JSON()
        request = MagicMock()
        request.getHeader.return_value = b'application/json'
        json_data = {'method': 'some.method', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        json._on_json_request(request)

    def test_on_json_request_valid_content_type_with_charset(self):
        """Ensure content-type parameters such as charset are ignored"""
        json = JSON()
        request = MagicMock()
        request.getHeader.return_value = b'application/json;charset=utf-8'
        json_data = {'method': 'some.method', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        json._on_json_request(request)


@pytest.mark.usefixtures('daemon', 'client', 'component')
class TestJSONCustomUserTestCase:
    @pytest_twisted.inlineCallbacks
    def test_handle_request_auth_error(self):
        json = JSON()
        auth_conf = {'session_timeout': 10, 'sessions': {}}
        Auth(auth_conf)  # Must create the component

        # Must be called to update remote methods in json object
        yield json.get_remote_methods()

        request = MagicMock()
        request.getCookie = MagicMock(return_value=b'bad_value')
        json_data = {'method': 'core.get_libtorrent_version', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        request_id, result, error = json._handle_request(request)
        assert error == {'message': 'Not authenticated', 'code': 1}


@pytest.mark.usefixtures('daemon', 'client', 'component')
class TestRPCRaiseDelugeErrorJSON:
    daemon_custom_script = """
    from deluge.error import DelugeError
    from deluge.core.rpcserver import export
    class TestClass(object):
        @export()
        def test(self):
            raise DelugeError('DelugeERROR')

    test = TestClass()
    daemon.rpcserver.register_object(test)
"""

    @pytest_twisted.ensureDeferred
    async def test_handle_request_method_raise_delugeerror(self):
        json = JSON()

        def get_session_id(s_id):
            return s_id

        self.patch(deluge.ui.web.auth, 'get_session_id', get_session_id)
        auth_conf = {'session_timeout': 10, 'sessions': {}}
        auth = Auth(auth_conf)
        request = Request(MagicMock(), False)
        request.base = b''
        auth._create_session(request)
        methods = await json.get_remote_methods()
        # Verify the function has been registered
        assert 'testclass.test' in methods

        request = MagicMock()
        session_id = list(auth.config['sessions'])[0]
        request.getCookie = MagicMock(return_value=session_id.encode())
        json_data = {'method': 'testclass.test', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        request_id, result, error = json._handle_request(request)
        with pytest.raises(DelugeError):
            await result


class TestJSONRequestFailed(WebServerMockBase):
    @pytest_twisted.async_yield_fixture(autouse=True)
    async def set_up(self, config_dir):
        custom_script = """
    from deluge.error import DelugeError
    from deluge.core.rpcserver import export
    from twisted.internet import reactor, task
    class TestClass(object):
        @export()
        def test(self):
            def test_raise_error():
                raise DelugeError('DelugeERROR')

            return task.deferLater(reactor, 1, test_raise_error)

    test = TestClass()
    daemon.rpcserver.register_object(test)
"""
        from twisted.internet.defer import Deferred

        extra_callback = {
            'deferred': Deferred(),
            'types': ['stderr'],
            'timeout': 10,
            'triggers': [
                {
                    'expr': 'in test_raise_error',
                    'value': lambda reader, data, data_all: 'Test',
                }
            ],
        }

        def on_test_raise(*args):
            assert 'Unhandled error in Deferred:' in self.core.stderr_out
            assert 'in test_raise_error' in self.core.stderr_out

        extra_callback['deferred'].addCallback(on_test_raise)
        d, daemon = common.start_core(
            custom_script=custom_script,
            print_stdout=False,
            print_stderr=False,
            timeout=5,
            extra_callbacks=[extra_callback],
            config_directory=config_dir,
        )
        await d
        yield
        await daemon.kill()

    @pytest_twisted.inlineCallbacks
    def test_render_on_rpc_request_failed(self, component, client):
        json = JSON()

        methods = yield json.get_remote_methods()
        # Verify the function has been registered
        assert 'testclass.test' in methods

        request = MagicMock()

        # Circumvent authentication
        auth = Auth({})
        self.mock_authentication_ignore(auth)

        def write(response_str):
            request.write_was_called = True
            response = json_lib.loads(response_str.decode())
            assert response['result'] is None, 'BAD RESULT'
            assert response['id'] == 0
            assert (
                response['error']['message']
                == 'Failure: [Failure instance: Traceback (failure with no frames):'
                " <class 'deluge.error.DelugeError'>: DelugeERROR\n]"
            )
            assert response['error']['code'] == 4

        request.write = write
        request.write_was_called = False
        request._disconnected = False
        request.getHeader.return_value = b'application/json'
        json_data = {'method': 'testclass.test', 'id': 0, 'params': []}
        request.json = json_lib.dumps(json_data).encode()
        d = json._on_json_request(request)

        def on_success(arg):
            assert arg == server.NOT_DONE_YET
            return True

        d.addCallbacks(on_success, pytest.fail)
        yield d



#
# Copyright (C) 2007-2010 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import traceback
from collections import defaultdict

from twisted.internet import reactor
from twisted.internet.defer import DeferredList, fail, maybeDeferred, succeed
from twisted.internet.task import LoopingCall, deferLater

log = logging.getLogger(__name__)


class ComponentAlreadyRegistered(Exception):
    pass


class ComponentException(Exception):
    def __init__(self, message, tb):
        super().__init__(message)
        self.message = message
        self.tb = tb

    def __str__(self):
        s = super().__str__()
        return '{}\n{}'.format(s, ''.join(self.tb))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.message == other.message
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class Component:
    """Component objects are singletons managed by the :class:`ComponentRegistry`.

    When a new Component object is instantiated, it will be automatically
    registered with the :class:`ComponentRegistry`.

    The ComponentRegistry has the ability to start, stop, pause and shutdown the
    components registered with it.

    **Events:**

        **start()** - This method is called when the client has connected to a
                  Deluge core.

        **stop()** - This method is called when the client has disconnected from a
                 Deluge core.

        **update()** - This method is called every 1 second by default while the
                   Componented is in a *Started* state.  The interval can be
                   specified during instantiation.  The update() timer can be
                   paused by instructing the :class:`ComponentRegistry` to pause
                   this Component.

        **shutdown()** - This method is called when the client is exiting.  If the
                     Component is in a "Started" state when this is called, a
                     call to stop() will be issued prior to shutdown().

    **States:**

        A Component can be in one of these 5 states.

        **Started** - The Component has been started by the :class:`ComponentRegistry`
                    and will have it's update timer started.

        **Starting** - The Component has had it's start method called, but it hasn't
                    fully started yet.

        **Stopped** - The Component has either been stopped or has yet to be started.

        **Stopping** - The Component has had it's stop method called, but it hasn't
                    fully stopped yet.

        **Paused** - The Component has had it's update timer stopped, but will
                    still be considered in a Started state.

    """

    def __init__(self, name, interval=1, depend=None):
        """Initialize component.

        Args:
            name (str): Name of component.
            interval (int, optional): The interval in seconds to call the update function.
            depend (list, optional): The names of components this component depends on.

        """
        self._component_name = name
        self._component_interval = interval
        self._component_depend = depend
        self._component_state = 'Stopped'
        self._component_timer = None
        self._component_starting_deferred = None
        self._component_stopping_deferred = None
        _ComponentRegistry.register(self)

    def __del__(self):
        if _ComponentRegistry:
            _ComponentRegistry.deregister(self)

    def _component_start_timer(self):
        if hasattr(self, 'update'):
            self._component_timer = LoopingCall(self.update)
            self._component_timer.start(self._component_interval)

    def _component_start(self):
        def on_start(result):
            self._component_state = 'Started'
            self._component_starting_deferred = None
            self._component_start_timer()
            return True

        def on_start_fail(result):
            self._component_state = 'Stopped'
            self._component_starting_deferred = None
            log.error(result)
            return fail(result)

        if self._component_state == 'Stopped':
            if hasattr(self, 'start'):
                self._component_state = 'Starting'
                d = deferLater(reactor, 0, self.start)
                d.addCallbacks(on_start, on_start_fail)
                self._component_starting_deferred = d
            else:
                d = maybeDeferred(on_start, None)
        elif self._component_state == 'Starting':
            return self._component_starting_deferred
        elif self._component_state == 'Started':
            d = succeed(True)
        else:
            d = fail(
                ComponentException(
                    'Trying to start component "%s" but it is '
                    'not in a stopped state. Current state: %s'
                    % (self._component_name, self._component_state),
                    traceback.format_stack(limit=4),
                )
            )
        return d

    def _component_stop(self):
        def on_stop(result):
            self._component_state = 'Stopped'
            if self._component_timer and self._component_timer.running:
                self._component_timer.stop()
            return True

        def on_stop_fail(result):
            self._component_state = 'Started'
            self._component_stopping_deferred = None
            log.error(result)
            return result

        if self._component_state != 'Stopped' and self._component_state != 'Stopping':
            if hasattr(self, 'stop'):
                self._component_state = 'Stopping'
                d = maybeDeferred(self.stop)
                d.addCallback(on_stop)
                d.addErrback(on_stop_fail)
                self._component_stopping_deferred = d
            else:
                d = maybeDeferred(on_stop, None)

        if self._component_state == 'Stopping':
            return self._component_stopping_deferred

        return succeed(None)

    def _component_pause(self):
        def on_pause(result):
            self._component_state = 'Paused'

        if self._component_state == 'Started':
            if self._component_timer and self._component_timer.running:
                d = maybeDeferred(self._component_timer.stop)
                d.addCallback(on_pause)
            else:
                d = succeed(None)
        elif self._component_state == 'Paused':
            d = succeed(None)
        else:
            d = fail(
                ComponentException(
                    'Trying to pause component "%s" but it is '
                    'not in a started state. Current state: %s'
                    % (self._component_name, self._component_state),
                    traceback.format_stack(limit=4),
                )
            )
        return d

    def _component_resume(self):
        def on_resume(result):
            self._component_state = 'Started'

        if self._component_state == 'Paused':
            d = maybeDeferred(self._component_start_timer)
            d.addCallback(on_resume)
        else:
            d = fail(
                ComponentException(
                    'Trying to resume component "%s" but it is '
                    'not in a paused state. Current state: %s'
                    % (self._component_name, self._component_state),
                    traceback.format_stack(limit=4),
                )
            )
        return d

    def _component_shutdown(self):
        def on_stop(result):
            if hasattr(self, 'shutdown'):
                return maybeDeferred(self.shutdown)
            return succeed(None)

        d = self._component_stop()
        d.addCallback(on_stop)
        return d

    def get_state(self):
        return self._component_state

    def start(self):
        pass

    def stop(self):
        pass

    def update(self):
        pass

    def shutdown(self):
        pass


class ComponentRegistry:
    """The ComponentRegistry holds a list of currently registered :class:`Component` objects.

    It is used to manage the Components by starting, stopping, pausing and shutting them down.
    """

    def __init__(self):
        self.components = {}
        # Stores all of the components that are dependent on a particular component
        self.dependents = defaultdict(list)

    def register(self, obj):
        """Register a component object with the registry.

        Note:
            This is done automatically when a Component object is instantiated.

        Args:
            obj (Component): A component object to register.

        Raises:
            ComponentAlreadyRegistered: If a component with the same name is already registered.

        """
        name = obj._component_name
        if name in self.components:
            raise ComponentAlreadyRegistered(
                'Component already registered with name %s' % name
            )

        self.components[obj._component_name] = obj
        if obj._component_depend:
            for depend in obj._component_depend:
                self.dependents[depend].append(name)

    def deregister(self, obj):
        """Deregister a component from the registry.  A stop will be
        issued to the component prior to deregistering it.

        Args:
            obj (Component): a component object to deregister

        Returns:
            Deferred: a deferred object that will fire once the Component has been
                successfully deregistered

        """
        if obj in self.components.values():
            log.debug('Deregistering Component: %s', obj._component_name)
            d = self.stop([obj._component_name])

            def on_stop(result, name):
                # Component may have been removed, so pop to ensure it doesn't fail
                self.components.pop(name, None)

            return d.addCallback(on_stop, obj._component_name)
        else:
            return succeed(None)

    def start(self, names=None):
        """Start Components, and their dependencies, that are currently in a Stopped state.

        Note:
            If no names are specified then all registered components will be started.

        Args:
            names (list): A list of Components to start and their dependencies.

        Returns:
            Deferred: Fired once all Components have been successfully started.

        """
        # Start all the components if names is empty
        if not names:
            names = list(self.components)
        elif isinstance(names, str):
            names = [names]

        def on_depends_started(result, name):
            return self.components[name]._component_start()

        deferreds = []

        for name in names:
            if self.components[name]._component_depend:
                # This component has depends, so we need to start them first.
                d = self.start(self.components[name]._component_depend)
                d.addCallback(on_depends_started, name)
                deferreds.append(d)
            else:
                deferreds.append(self.components[name]._component_start())

        return DeferredList(deferreds)

    def stop(self, names=None):
        """Stop Components that are currently not in a Stopped state.

        Note:
            If no names are specified then all registered components will be stopped.

        Args:
            names (list): A list of Components to stop.

        Returns:
            Deferred: Fired once all Components have been successfully stopped.

        """
        if not names:
            names = list(self.components)
        elif isinstance(names, str):
            names = [names]

        def on_dependents_stopped(result, name):
            return self.components[name]._component_stop()

        stopped_in_deferred = set()
        deferreds = []

        for name in names:
            if name in stopped_in_deferred:
                continue
            if name in self.components:
                if name in self.dependents:
                    # If other components depend on this component, stop them first
                    d = self.stop(self.dependents[name]).addCallback(
                        on_dependents_stopped, name
                    )
                    deferreds.append(d)
                    stopped_in_deferred.update(self.dependents[name])
                else:
                    deferreds.append(self.components[name]._component_stop())

        return DeferredList(deferreds)

    def pause(self, names=None):
        """Pause Components that are currently in a Started state.

        Note:
            If no names are specified then all registered components will be paused.

        Args:
            names (list): A list of Components to pause.

        Returns:
            Deferred: Fired once all Components have been successfully paused.

        """
        if not names:
            names = list(self.components)
        elif isinstance(names, str):
            names = [names]

        deferreds = []

        for name in names:
            if self.components[name]._component_state == 'Started':
                deferreds.append(self.components[name]._component_pause())

        return DeferredList(deferreds)

    def resume(self, names=None):
        """Resume Components that are currently in a Paused state.

        Note:
            If no names are specified then all registered components will be resumed.

        Args:
            names (list): A list of Components to to resume.

        Returns:
            Deferred: Fired once all Components have been successfully resumed.

        """
        if not names:
            names = list(self.components)
        elif isinstance(names, str):
            names = [names]

        deferreds = []

        for name in names:
            if self.components[name]._component_state == 'Paused':
                deferreds.append(self.components[name]._component_resume())

        return DeferredList(deferreds)

    def shutdown(self):
        """Shutdown all Components regardless of state.

        This will call stop() on all the components prior to shutting down. This should be called
        when the program is exiting to ensure all Components have a chance to properly shutdown.

        Returns:
            Deferred: Fired once all Components have been successfully shut down.

        """

        def on_stopped(result):
            return DeferredList(
                [comp._component_shutdown() for comp in list(self.components.values())]
            )

        return self.stop(list(self.components)).addCallback(on_stopped)

    def update(self):
        """Update all Components that are in a Started state."""
        for component in self.components.items():
            try:
                component.update()
            except BaseException as ex:
                log.exception(ex)


_ComponentRegistry = ComponentRegistry()

deregister = _ComponentRegistry.deregister
start = _ComponentRegistry.start
stop = _ComponentRegistry.stop
pause = _ComponentRegistry.pause
resume = _ComponentRegistry.resume
update = _ComponentRegistry.update
shutdown = _ComponentRegistry.shutdown


def get(name):
    """Return a reference to a component.

    Args:
        name (str): The Component name to get.

    Returns:
        Component: The Component object.

    Raises:
        KeyError: If the Component does not exist.

    """
    return _ComponentRegistry.components[name]



#!/usr/bin/env python
#
# Copyright (C) 2013 Bro <bro.development@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os


def is_hidden(filepath):
    def has_hidden_attribute(filepath):
        import win32api
        import win32con

        try:
            attribute = win32api.GetFileAttributes(filepath)
            return attribute & (
                win32con.FILE_ATTRIBUTE_HIDDEN | win32con.FILE_ATTRIBUTE_SYSTEM
            )
        except (AttributeError, AssertionError):
            return False

    name = os.path.basename(os.path.abspath(filepath))
    # Windows
    if os.name == 'nt':
        return has_hidden_attribute(filepath)
    return name.startswith('.')


def get_completion_paths(args):
    """
    Takes a path value and returns the available completions.
    If the path_value is a valid path, return all sub-directories.
    If the path_value is not a valid path, remove the basename from the
    path and return all sub-directories of path that start with basename.

    :param args: options
    :type args: dict
    :returns: the args argument containing the available completions for the completion_text
    :rtype: dict

    """
    args['paths'] = []
    path_value = args['completion_text']
    hidden_files = args['show_hidden_files']

    def get_subdirs(dirname):
        try:
            return next(os.walk(dirname))[1]
        except StopIteration:
            # Invalid dirname
            return []

    dirname = os.path.dirname(path_value)
    basename = os.path.basename(path_value)

    dirs = get_subdirs(dirname)
    # No completions available
    if not dirs:
        return args

    # path_value ends with path separator so
    # we only want all the subdirectories
    if not basename:
        # Lets remove hidden files
        if not hidden_files:
            old_dirs = dirs
            dirs = []
            for d in old_dirs:
                if not is_hidden(os.path.join(dirname, d)):
                    dirs.append(d)
    matching_dirs = []
    for s in dirs:
        if s.startswith(basename):
            p = os.path.join(dirname, s)
            if not p.endswith(os.path.sep):
                p += os.path.sep
            matching_dirs.append(p)

    args['paths'] = sorted(matching_dirs)
    return args



#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.decorators import overrides
from deluge.ui.console.modes.basemode import BaseMode
from deluge.ui.console.utils import curses_util as util

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


class EventView(BaseMode):
    def __init__(self, parent_mode, stdscr, encoding=None):
        BaseMode.__init__(self, stdscr, encoding)
        self.parent_mode = parent_mode
        self.offset = 0

    def back_to_overview(self):
        component.get('ConsoleUI').set_mode(self.parent_mode.mode_name)

    @overrides(component.Component)
    def update(self):
        self.refresh()

    @overrides(BaseMode)
    def refresh(self):
        """
        This method just shows each line of the event log
        """
        events = component.get('ConsoleUI').events

        self.stdscr.erase()
        self.draw_statusbars()

        if events:
            for i, event in enumerate(events):
                if i - self.offset >= self.rows - 2:
                    more = len(events) - self.offset - self.rows + 2
                    if more > 0:
                        self.add_string(i - self.offset, '  (And %i more)' % more)
                    break

                elif i - self.offset < 0:
                    continue
                try:
                    self.add_string(i + 1 - self.offset, event)
                except curses.error:
                    pass  # This'll just cut the line. Note: This seriously should be fixed in a better way
        else:
            self.add_string(1, '{!white,black,bold!}No events to show yet')

        if not component.get('ConsoleUI').is_active_mode(self):
            return

        self.stdscr.noutrefresh()
        curses.doupdate()

    @overrides(BaseMode)
    def on_resize(self, rows, cols):
        BaseMode.on_resize(self, rows, cols)
        self.refresh()

    @overrides(BaseMode)
    def read_input(self):
        c = self.stdscr.getch()

        if c in [ord('q'), util.KEY_ESC]:
            self.back_to_overview()
            return

        # TODO: Scroll event list
        jumplen = self.rows - 3
        num_events = len(component.get('ConsoleUI').events)

        if c == curses.KEY_UP:
            self.offset -= 1
        elif c == curses.KEY_PPAGE:
            self.offset -= jumplen
        elif c == curses.KEY_HOME:
            self.offset = 0
        elif c == curses.KEY_DOWN:
            self.offset += 1
        elif c == curses.KEY_NPAGE:
            self.offset += jumplen
        elif c == curses.KEY_END:
            self.offset += num_events
        elif c == ord('j'):
            self.offset += 1
        elif c == ord('k'):
            self.offset -= 1

        if self.offset <= 0:
            self.offset = 0
        elif num_events > self.rows - 3:
            if self.offset > num_events - self.rows + 3:
                self.offset = num_events - self.rows + 3
        else:
            self.offset = 0

        self.refresh()



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import argparse
import sys
from io import StringIO
from unittest import mock

import pytest
import pytest_twisted
from twisted.internet import defer

import deluge
import deluge.component as component
import deluge.ui.console
import deluge.ui.console.cmdline.commands.quit
import deluge.ui.console.main
import deluge.ui.web.server
from deluge.common import get_localhost_auth, windows_check
from deluge.conftest import BaseTestCase
from deluge.ui import ui_entry
from deluge.ui.web.server import DelugeWeb

from . import common
from .daemon_base import DaemonBase

DEBUG_COMMAND = False

sys_stdout = sys.stdout
# To catch output to stdout/stderr while running unit tests, we patch
# the file descriptors in sys and argparse._sys with StringFileDescriptor.
# Regular print statements from such tests will therefore write to the
# StringFileDescriptor object instead of the terminal.
# To print to terminal from the tests, use: print('Message...', file=sys_stdout)


class StringFileDescriptor:
    """File descriptor that writes to string buffer"""

    def __init__(self, fd):
        self.out = StringIO()
        self.fd = fd
        for a in ['encoding']:
            setattr(self, a, getattr(sys_stdout, a))

    def write(self, *data, **kwargs):
        data_string = str(*data)
        print(data_string, file=self.out, end='')

    def flush(self):
        self.out.flush()


class UIBaseTestCase:
    def set_up(self):
        common.setup_test_logger(level='info', prefix=self.id())
        return component.start()

    def tear_down(self):
        return component.shutdown()

    def exec_command(self):
        if DEBUG_COMMAND:
            print('Executing: %s\n' % sys.argv, file=sys_stdout)
        return self.var['start_cmd']()


class UIWithDaemonBaseTestCase(UIBaseTestCase, DaemonBase):
    """Subclass for test that require a deluged daemon"""

    def set_up(self):
        d = self.common_set_up()
        common.setup_test_logger(level='info', prefix=self.id())
        return d


class TestDelugeEntry(BaseTestCase):
    def set_up(self):
        return component.start()

    def tear_down(self):
        return component.shutdown()

    def test_deluge_help(self):
        self.patch(sys, 'argv', ['./deluge', '-h'])
        config = deluge.configmanager.ConfigManager('ui.conf', ui_entry.DEFAULT_PREFS)
        config.config['default_ui'] = 'console'
        config.save()

        fd = StringFileDescriptor(sys.stdout)
        self.patch(argparse._sys, 'stdout', fd)

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            with pytest.raises(SystemExit):
                ui_entry.start_ui()
            assert 'usage: deluge' in fd.out.getvalue()
            assert 'UI Options:' in fd.out.getvalue()
            assert '* console' in fd.out.getvalue()

    def test_start_default(self):
        self.patch(sys, 'argv', ['./deluge'])
        config = deluge.configmanager.ConfigManager('ui.conf', ui_entry.DEFAULT_PREFS)
        config.config['default_ui'] = 'console'
        config.save()

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            # Just test that no exception is raised
            ui_entry.start_ui()

    def test_start_with_log_level(self):
        _level = []

        def setup_logger(
            level='error',
            filename=None,
            filemode='w',
            logrotate=None,
            output_stream=sys.stdout,
        ):
            _level.append(level)

        self.patch(deluge.log, 'setup_logger', setup_logger)
        self.patch(sys, 'argv', ['./deluge', '-L', 'info'])

        config = deluge.configmanager.ConfigManager('ui.conf', ui_entry.DEFAULT_PREFS)
        config.config['default_ui'] = 'console'
        config.save()

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            # Just test that no exception is raised
            ui_entry.start_ui()

        assert _level[0] == 'info'


class GtkUIBaseTestCase(UIBaseTestCase):
    """Implement all GtkUI tests here"""

    def test_start_gtk3ui(self):
        self.patch(sys, 'argv', self.var['sys_arg_cmd'])

        from deluge.ui.gtk3 import gtkui

        with mock.patch.object(gtkui.GtkUI, 'start', autospec=True):
            self.exec_command()


@pytest.mark.gtkui
class TestGtkUIDelugeScriptEntry(BaseTestCase, GtkUIBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        request.cls.var = {
            'cmd_name': 'deluge gtk',
            'start_cmd': ui_entry.start_ui,
            'sys_arg_cmd': ['./deluge', 'gtk'],
        }


@pytest.mark.gtkui
class TestGtkUIScriptEntry(BaseTestCase, GtkUIBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        from deluge.ui import gtk3

        request.cls.var = {
            'cmd_name': 'deluge-gtk',
            'start_cmd': gtk3.start,
            'sys_arg_cmd': ['./deluge-gtk'],
        }


class DelugeWebMock(DelugeWeb):
    def __init__(self, *args, **kwargs):
        kwargs['daemon'] = False
        DelugeWeb.__init__(self, *args, **kwargs)


class WebUIBaseTestCase(UIBaseTestCase):
    """Implement all WebUI tests here"""

    def test_start_webserver(self):
        self.patch(sys, 'argv', self.var['sys_arg_cmd'])
        self.patch(deluge.ui.web.server, 'DelugeWeb', DelugeWebMock)
        self.exec_command()

    def test_start_web_with_log_level(self):
        _level = []

        def setup_logger(
            level='error',
            filename=None,
            filemode='w',
            logrotate=None,
            output_stream=sys.stdout,
        ):
            _level.append(level)

        self.patch(deluge.log, 'setup_logger', setup_logger)
        self.patch(sys, 'argv', self.var['sys_arg_cmd'] + ['-L', 'info'])

        config = deluge.configmanager.ConfigManager('ui.conf', ui_entry.DEFAULT_PREFS)
        config.config['default_ui'] = 'web'
        config.save()

        self.patch(deluge.ui.web.server, 'DelugeWeb', DelugeWebMock)
        self.exec_command()
        assert _level[0] == 'info'


class TestWebUIScriptEntry(BaseTestCase, WebUIBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        request.cls.var = {
            'cmd_name': 'deluge-web',
            'start_cmd': deluge.ui.web.start,
            'sys_arg_cmd': ['./deluge-web'],
        }
        if not windows_check():
            request.cls.var['sys_arg_cmd'].append('--do-not-daemonize')


class TestWebUIDelugeScriptEntry(BaseTestCase, WebUIBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        request.cls.var = {
            'cmd_name': 'deluge web',
            'start_cmd': ui_entry.start_ui,
            'sys_arg_cmd': ['./deluge', 'web'],
        }
        if not windows_check():
            request.cls.var['sys_arg_cmd'].append('--do-not-daemonize')


class ConsoleUIBaseTestCase(UIBaseTestCase):
    """Implement Console tests that do not require a running daemon"""

    def test_start_console(self):
        self.patch(sys, 'argv', self.var['sys_arg_cmd'])
        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            self.exec_command()

    def test_start_console_with_log_level(self, request):
        _level = []

        def setup_logger(
            level='error',
            filename=None,
            filemode='w',
            logrotate=None,
            output_stream=sys.stdout,
        ):
            _level.append(level)

        self.patch(deluge.log, 'setup_logger', setup_logger)
        self.patch(sys, 'argv', self.var['sys_arg_cmd'] + ['-L', 'info'])

        config = deluge.configmanager.ConfigManager('ui.conf', ui_entry.DEFAULT_PREFS)
        config.config['default_ui'] = 'console'
        config.save()

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            # Just test that no exception is raised
            self.exec_command()

        assert _level[0] == 'info'

    def test_console_help(self):
        self.patch(sys, 'argv', self.var['sys_arg_cmd'] + ['-h'])
        fd = StringFileDescriptor(sys.stdout)
        self.patch(argparse._sys, 'stdout', fd)

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            with pytest.raises(SystemExit):
                self.exec_command()
            std_output = fd.out.getvalue()
            assert (
                'usage: %s' % self.var['cmd_name']
            ) in std_output  # Check command name
            assert 'Common Options:' in std_output
            assert 'Console Options:' in std_output
            assert (
                'Console Commands:\n  The following console commands are available:'
                in std_output
            )
            assert 'The following console commands are available:' in std_output

    def test_console_command_info(self):
        self.patch(sys, 'argv', self.var['sys_arg_cmd'] + ['info'])
        fd = StringFileDescriptor(sys.stdout)
        self.patch(argparse._sys, 'stdout', fd)

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            self.exec_command()

    def test_console_command_info_help(self):
        self.patch(sys, 'argv', self.var['sys_arg_cmd'] + ['info', '-h'])
        fd = StringFileDescriptor(sys.stdout)
        self.patch(argparse._sys, 'stdout', fd)

        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            with pytest.raises(SystemExit):
                self.exec_command()
            std_output = fd.out.getvalue()
            assert 'usage: info' in std_output
            assert 'Show information about the torrents' in std_output

    def test_console_unrecognized_arguments(self):
        self.patch(
            sys, 'argv', ['./deluge', '--ui', 'console']
        )  # --ui is not longer supported
        fd = StringFileDescriptor(sys.stdout)
        self.patch(argparse._sys, 'stderr', fd)
        with mock.patch('deluge.ui.console.main.ConsoleUI'):
            with pytest.raises(SystemExit):
                self.exec_command()
            assert 'unrecognized arguments: --ui' in fd.out.getvalue()


class ConsoleUIWithDaemonBaseTestCase(UIWithDaemonBaseTestCase):
    """Implement Console tests that require a running daemon"""

    def set_up(self):
        # Avoid calling reactor.shutdown after commands are executed by main.exec_args()
        deluge.ui.console.main.reactor = common.ReactorOverride()
        return UIWithDaemonBaseTestCase.set_up(self)

    def patch_arg_command(self, command):
        if type(command) == str:
            command = [command]
        username, password = get_localhost_auth()
        self.patch(
            sys,
            'argv',
            self.var['sys_arg_cmd']
            + ['--port']
            + [str(self.listen_port)]
            + ['--username']
            + [username]
            + ['--password']
            + [password]
            + command,
        )

    @pytest_twisted.inlineCallbacks
    def test_console_command_add(self):
        filename = common.get_test_data_file('test.torrent')
        self.patch_arg_command([f'add "{filename}"'])
        fd = StringFileDescriptor(sys.stdout)
        self.patch(sys, 'stdout', fd)

        yield self.exec_command()

        std_output = fd.out.getvalue()
        assert (
            std_output
            == 'Attempting to add torrent: ' + filename + '\nTorrent added!\n'
        )

    @pytest_twisted.inlineCallbacks
    def test_console_command_add_move_completed(self):
        filename = common.get_test_data_file('test.torrent')
        tmp_path = 'c:\\tmp' if windows_check() else '/tmp'
        self.patch_arg_command(
            [
                f'add --move-path "{tmp_path}" "{filename}" ; status'
                ' ; manage'
                ' ab570cdd5a17ea1b61e970bb72047de141bce173'
                ' move_completed'
                ' move_completed_path'
            ]
        )
        fd = StringFileDescriptor(sys.stdout)
        self.patch(sys, 'stdout', fd)

        yield self.exec_command()

        std_output = fd.out.getvalue()
        assert std_output.endswith(
            f'move_completed: True\nmove_completed_path: {tmp_path}\n'
        ) or std_output.endswith(
            f'move_completed_path: {tmp_path}\nmove_completed: True\n'
        )

    @pytest_twisted.ensureDeferred
    async def test_console_command_status(self):
        fd = StringFileDescriptor(sys.stdout)
        self.patch_arg_command(['status'])
        self.patch(sys, 'stdout', fd)

        await self.exec_command()

        std_output = fd.out.getvalue()
        assert std_output.startswith('Total upload: ')
        assert std_output.endswith(' Moving: 0\n')

    @defer.inlineCallbacks
    def test_console_command_config_set_download_location(self):
        fd = StringFileDescriptor(sys.stdout)
        self.patch_arg_command(['config --set download_location /downloads'])
        self.patch(sys, 'stdout', fd)

        yield self.exec_command()
        std_output = fd.out.getvalue()
        assert std_output.startswith('Setting "download_location" to: \'/downloads\'')
        assert std_output.endswith('Configuration value successfully updated.\n')


@pytest.mark.usefixtures('daemon', 'client')
class TestConsoleScriptEntryWithDaemon(BaseTestCase, ConsoleUIWithDaemonBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        request.cls.var = {
            'cmd_name': 'deluge-console',
            'start_cmd': deluge.ui.console.start,
            'sys_arg_cmd': ['./deluge-console'],
        }


class TestConsoleScriptEntry(BaseTestCase, ConsoleUIBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        request.cls.var = {
            'cmd_name': 'deluge-console',
            'start_cmd': deluge.ui.console.start,
            'sys_arg_cmd': ['./deluge-console'],
        }


class TestConsoleDelugeScriptEntry(BaseTestCase, ConsoleUIBaseTestCase):
    @pytest.fixture(autouse=True)
    def set_var(self, request):
        request.cls.var = {
            'cmd_name': 'deluge console',
            'start_cmd': ui_entry.start_ui,
            'sys_arg_cmd': ['./deluge', 'console'],
        }



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import pytest

from deluge import bencode

from . import common


class TestBencode:
    def test_bencode_unicode_metainfo(self):
        filename = common.get_test_data_file('test.torrent')
        with open(filename, 'rb') as _file:
            metainfo = bencode.bdecode(_file.read())[b'info']
        bencode.bencode({b'info': metainfo})

    def test_bencode_unicode_value(self):
        assert bencode.bencode(b'abc') == b'3:abc'
        assert bencode.bencode('abc') == b'3:abc'

    def test_bdecode(self):
        assert bencode.bdecode(b'3:dEf') == b'dEf'
        with pytest.raises(bencode.BTFailure):
            bencode.bdecode('dEf')
        with pytest.raises(bencode.BTFailure):
            bencode.bdecode(b'dEf')
        with pytest.raises(bencode.BTFailure):
            bencode.bdecode({'dEf': 123})



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import deluge.component as component
from deluge.ui.client import client

from . import BaseCommand


class Command(BaseCommand):
    """Shutdown the deluge server."""

    def handle(self, options):
        self.console = component.get('ConsoleUI')

        def on_shutdown(result):
            self.console.write('{!success!}Daemon was shutdown')

        def on_shutdown_fail(reason):
            self.console.write('{!error!}Unable to shutdown daemon: %s' % reason)

        return (
            client.daemon.shutdown()
            .addCallback(on_shutdown)
            .addErrback(on_shutdown_fail)
        )



#
# This file is public domain.
#

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'


# Deferred translation
def _(message):
    return message


# Languages we provide translations for, out of the box.
LANGUAGES = {
    'af': _('Afrikaans'),
    'ar': _('Arabic'),
    'ast': _('Asturian'),
    'az': _('Azerbaijani'),
    'bg': _('Bulgarian'),
    'be': _('Belarusian'),
    'bn': _('Bengali'),
    'br': _('Breton'),
    'bs': _('Bosnian'),
    'ca': _('Catalan'),
    'cs': _('Czech'),
    'cy': _('Welsh'),
    'da': _('Danish'),
    'de': _('German'),
    'el': _('Greek'),
    'en': _('English'),
    'en_AU': _('English (Australia)'),
    'en_CA': _('English (Canada)'),
    'en_GB': _('English (United Kingdom)'),
    'eo': _('Esperanto'),
    'es': _('Spanish'),
    'es-ar': _('Argentinian Spanish'),
    'es-mx': _('Mexican Spanish'),
    'es-ni': _('Nicaraguan Spanish'),
    'es-ve': _('Venezuelan Spanish'),
    'et': _('Estonian'),
    'eu': _('Basque'),
    'fa': _('Persian'),
    'fi': _('Finnish'),
    'fr': _('French'),
    'fy': _('Frisian'),
    'ga': _('Irish'),
    'gl': _('Galician'),
    'he': _('Hebrew'),
    'hi': _('Hindi'),
    'hr': _('Croatian'),
    'hu': _('Hungarian'),
    'ia': _('Interlingua'),
    'id': _('Indonesian'),
    'is': _('Icelandic'),
    'it': _('Italian'),
    'iu': _('Inuktitut'),
    'ja': _('Japanese'),
    'ka': _('Georgian'),
    'kk': _('Kazakh'),
    'km': _('Khmer'),
    'kn': _('Kannada'),
    'ko': _('Korean'),
    'ku': _('Kurdish'),
    'la': _('Latin'),
    'lb': _('Luxembourgish'),
    'lt': _('Lithuanian'),
    'lv': _('Latvian'),
    'mk': _('Macedonian'),
    'ml': _('Malayalam'),
    'mn': _('Mongolian'),
    'ms': _('Mayaly'),
    'my': _('Burmese'),
    'nb': _('Norwegian Bokmal'),
    'ne': _('Nepali'),
    'nds': _('Low German'),
    'nl': _('Dutch'),
    'nn': _('Norwegian Nynorsk'),
    'os': _('Ossetic'),
    'pa': _('Punjabi'),
    'pl': _('Polish'),
    'pms': _('Piedmontese'),
    'pt': _('Portuguese'),
    'pt_BR': _('Brazilian Portuguese'),
    'ro': _('Romanian'),
    'ru': _('Russian'),
    'sk': _('Slovak'),
    'sl': _('Slovenian'),
    'si': _('Sinhalese'),
    'sq': _('Albanian'),
    'sr': _('Serbian'),
    'sr-latn': _('Serbian Latin'),
    'sv': _('Swedish'),
    'sw': _('Swahili'),
    'ta': _('Tamil'),
    'te': _('Telugu'),
    'th': _('Thai'),
    'tl': _('Tagalog'),
    'tlh': _('Klingon'),
    'tr': _('Turkish'),
    'tt': _('Tatar'),
    'udm': _('Udmurt'),
    'uk': _('Ukrainian'),
    'ur': _('Urdu'),
    'vi': _('Vietnamese'),
    'zh_CN': _('Chinese (Simplified)'),
    'zh_HK': _('Chinese (Hong Kong)'),
    'zh-hans': _('Simplified Chinese'),
    'zh-hant': _('Traditional Chinese'),
    'zh_TW': _('Chinese (Taiwan)'),
}

del _



#!/usr/bin/env python
#
# Copyright (C) 2009-2012 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""Script to parse javascript files for translation strings and generate gettext.js"""

import os
import re

WEBUI_JS_DIR = 'deluge/ui/web/js/deluge-all'
# Enabling Debug adds file and line number as comments to the gettext file.
DEBUG = False


def check_missing_markup(js_dir):
    """Search js to check for missed translation markup."""

    # A list of common extjs attributes that are usually marked for translation.
    attr_list = [
        "text: '",
        "msg: '",
        "title: '",
        "fieldLabel: '",
        "boxLabel: '",
        "tooltip: '",
        "header: '",
        "defaultText: '",
        "unit: '",
        r"setText\('",
        r"addButton\('",
    ]

    # Don't match against any of these chars at start of string value.
    except_chars = "' &#"

    # A list of strings that should be skipped shuold the match contain them.
    skip = ['HTTP:']

    # Create a list of the matching strings to search for with the except_chars appended to each one.
    string_re = re.compile(
        '('
        + ')|('.join(['%s[^' + except_chars + "].*'"] * len(attr_list))
        % tuple(attr_list)
        + ')'
    )

    strings = {}
    for root, dnames, files in os.walk(js_dir):
        for filename in files:
            if os.path.splitext(filename)[1] != '.js':
                continue
            for lineno, line in enumerate(open(os.path.join(root, filename))):
                for match in string_re.finditer(line):
                    for string in match.groups():
                        # Ignore string that contains only digits or specificied strings in skip.
                        if (
                            not string
                            or string.split('\'')[1].isdigit()
                            or any(x in string for x in skip)
                        ):
                            continue
                        locations = strings.get(string, [])
                        locations.append(
                            (os.path.join(root, filename), str(lineno + 1))
                        )
                        strings[string] = locations
    return strings


GETTEXT_TPL = (
    'GetText={maps:{},'
    'add:function(string,translation){this.maps[string]=translation},'
    'get:function(string){if (this.maps[string]){string=this.maps[string]} return string}};'
    'function _(string){return GetText.get(string)}'
)
GETTEXT_SUBST_TPL = "GetText.add('{key}','${{escape(_(\"{key}\"))}}')\n"


def create_gettext_js(js_dir):
    string_re = re.compile('_\\(\'(.*?)\'\\)')
    strings = {}
    for root, dnames, files in os.walk(js_dir):
        for filename in files:
            if filename.endswith('.js'):
                filepath = os.path.join(root, filename)
                with open(os.path.join(root, filename)) as _file:
                    for lineno, line in enumerate(_file, 1):
                        for match in string_re.finditer(line):
                            string = match.group(1)
                            locations = strings.get(string, [])
                            locations.append((filepath, lineno))
                            strings[string] = locations

    gettext_file = os.path.join(os.path.dirname(js_dir), 'gettext.js')
    with open(gettext_file, 'w') as fp:
        fp.write(GETTEXT_TPL)
        for key in sorted(strings):
            if DEBUG:
                fp.write(
                    '\n//: %s' % '//: '.join(['%s:%s\n' % x for x in strings[key]])
                )
            fp.write(GETTEXT_SUBST_TPL.format(key=key))
    return gettext_file


if __name__ == '__main__':
    gettext_fname = create_gettext_js(WEBUI_JS_DIR)
    print('Created: %s' % gettext_fname)
    missed_markup = check_missing_markup(WEBUI_JS_DIR)
    if missed_markup:
        print('Possible missed text for translation markup:')
        for text, filenames in missed_markup.iteritems():
            for filename_lineno in filenames:
                print('{:<58}    {}'.format(':'.join(filename_lineno), text))



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge import component  # for systray
from deluge.plugins.pluginbase import Gtk3PluginBase

from . import label_config, sidebar_menu, submenu

log = logging.getLogger(__name__)

NO_LABEL = 'No Label'


def cell_data_label(column, cell, model, row, data):
    cell.set_property('text', str(model.get_value(row, data)))


class GtkUI(Gtk3PluginBase):
    def start(self):
        if self.label_menu:
            self.label_menu.on_show()

    def enable(self):
        self.plugin = component.get('PluginManager')
        self.torrentmenu = component.get('MenuBar').torrentmenu
        self.label_menu = None
        self.labelcfg = None
        self.sidebar_menu = None
        self.load_interface()

    def disable(self):
        if self.label_menu in self.torrentmenu.get_children():
            self.torrentmenu.remove(self.label_menu)

        self.labelcfg.unload()
        self.sidebar_menu.unload()
        del self.sidebar_menu

        component.get('TorrentView').remove_column(_('Label'))

    def load_interface(self):
        # sidebar
        # disabled
        if not self.sidebar_menu:
            self.sidebar_menu = sidebar_menu.LabelSidebarMenu()
        # self.sidebar.load()

        # menu:
        log.debug('add items to torrentview-popup menu.')
        self.label_menu = submenu.LabelMenu()
        self.torrentmenu.append(self.label_menu)
        self.label_menu.show_all()

        # columns:
        self.load_columns()

        # config:
        if not self.labelcfg:
            self.labelcfg = label_config.LabelConfig(self.plugin)
        self.labelcfg.load()

        log.debug('Finished loading Label plugin')

    def load_columns(self):
        log.debug('add columns')

        component.get('TorrentView').add_text_column(_('Label'), status_field=['label'])



#!/usr/bin/env python
#
# Copyright (C) 2014 Calum Lind <calumlind@gmail.com>
# Copyright (C) 2010 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

"""Minifies the WebUI JS files.

Usage: python minify_web_js.py deluge/ui/web/js/deluge-all

"""

import fileinput
import fnmatch
import os
import subprocess
import sys
from shutil import which

closure_cmd = None
for cmd in ['closure-compiler', 'closure']:
    if which(cmd):
        closure_cmd = cmd
        break


def minify_closure(file_in, file_out):
    try:
        subprocess.check_call(
            [
                closure_cmd,
                '--warning_level',
                'QUIET',
                '--language_in=ECMASCRIPT5',
                '--js',
                file_in,
                '--js_output_file',
                file_out,
            ]
        )
        return True
    except subprocess.CalledProcessError:
        return False


# Closure outputs smallest files but java-based command, can use rJSmin
# as a python-only fallback.
#
#   deluge-all.js: Closure 131K, rJSmin: 148K
#
if not closure_cmd:
    try:
        from rjsmin import jsmin as minify
    except ImportError:
        print('Warning: No minifying command found.')
        minify = None


def source_files_list(source_dir):
    scripts = []
    for root, dirnames, filenames in os.walk(source_dir):
        dirnames.sort(reverse=True)
        files = fnmatch.filter(filenames, '*.js')
        files.sort()

        order_file = os.path.join(root, '.order')
        if os.path.isfile(order_file):
            with open(order_file) as _file:
                for line in _file:
                    if line.startswith('+ '):
                        order_filename = line.split()[1]
                        files.pop(files.index(order_filename))
                        files.insert(0, order_filename)

        # Ensure root directory files are bottom of list.
        if dirnames:
            scripts.extend([os.path.join(root, f) for f in files])
        else:
            for filename in reversed(files):
                scripts.insert(0, os.path.join(root, filename))
    return scripts


def concat_src_files(file_list, fileout_path):
    with open(fileout_path, 'w') as file_out:
        file_in = fileinput.input(file_list)
        file_out.writelines(file_in)


def minify_file(file_debug, file_minified):
    if closure_cmd:
        return minify_closure(file_debug, file_minified)
    elif minify:
        with open(file_minified, 'w') as file_out:
            with open(file_debug) as file_in:
                file_out.write(minify(file_in.read()))
                return True


def minify_js_dir(source_dir):
    build_name = os.path.basename(source_dir)
    build_dir = os.path.dirname(source_dir)
    file_debug_js = os.path.join(build_dir, build_name + '-debug.js')
    file_minified_js = os.path.join(build_dir, build_name + '.js')
    source_files = source_files_list(source_dir)

    if not source_files:
        print('No js files found, skipping %s' % source_dir)
        return

    concat_src_files(source_files, file_debug_js)
    print('Minifying %s' % source_dir)
    if not minify_file(file_debug_js, file_minified_js):
        print('Warning: Failed minifying files %s, debug only' % source_dir)
        if os.path.isfile(file_minified_js):
            os.remove(file_minified_js)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        JS_SOURCE_DIRS = [
            'deluge/ui/web/js/deluge-all',
            'deluge/ui/web/js/extjs/ext-extensions',
        ]
    else:
        JS_SOURCE_DIRS = [os.path.abspath(sys.argv[1])]

    for js_source_dir in JS_SOURCE_DIRS:
        minify_js_dir(js_source_dir)



#!/usr/bin/env python
# Authors: Douglas Creager <dcreager@dcreager.net>
#          Calum Lind <calumlind@gmail.com>
#
# This file is placed into the public domain.
#
# Calculates the current version number by first checking output of “git describe”,
# modified to conform to PEP 386 versioning scheme.  If “git describe” fails
# (likely due to using release tarball rather than git working copy), then fall
# back on reading the contents of the RELEASE-VERSION file.
#
# Usage: Import in setup.py, and use result of get_version() as package version:
#
# from version import *
#
# setup(
#     ...
#     version=get_version(),
#     ...
# )
#
# Script will automatically update the RELEASE-VERSION file, if needed.
# Note that  RELEASE-VERSION file should *not* be checked into git; please add
# it to your top-level .gitignore file.
#
# You'll probably want to distribute the RELEASE-VERSION file in your
# sdist tarballs; to do this, just create a MANIFEST.in file that
# contains the following line:
#
#   include RELEASE-VERSION
#

import os
import subprocess

__all__ = ('get_version',)

VERSION_FILE = os.path.join(os.path.dirname(__file__), 'RELEASE-VERSION')


def call_git_describe(prefix='', suffix=''):
    cmd = 'git describe --tags --match %s[0-9]*' % prefix
    try:
        output = subprocess.check_output(cmd.split(), stderr=subprocess.PIPE)
    except (OSError, subprocess.CalledProcessError):
        return None
    else:
        version = output.decode('utf-8').strip().replace(prefix, '')
        # A dash signifies git commit increments since parent tag.
        if '-' in version:
            segment = '.dev' if 'dev' in version else '.post'
            version = segment.join(version.replace(suffix, '').split('-')[:2])
        return version


def get_version(prefix='deluge-', suffix='.dev0'):
    try:
        with open(VERSION_FILE) as f:
            release_version = f.readline().strip()
    except OSError:
        release_version = None

    version = call_git_describe(prefix, suffix)

    if not version:
        version = release_version
    if not version:
        raise ValueError('Cannot find the version number!')

    if version != release_version:
        with open(VERSION_FILE, 'w') as f:
            f.write('%s\n' % version)

    return version


if __name__ == '__main__':
    print(get_version())



#
# Copyright (C) 2014 Bro <bro.development@gmail.com>
# Copyright (C) 2014 Calum Lind <calumlind@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import pytest

import deluge.component as component
from deluge.configmanager import ConfigManager
from deluge.conftest import BaseTestCase
from deluge.i18n import setup_translation

# Allow running other tests without GTKUI dependencies available
try:
    # pylint: disable=ungrouped-imports
    from gi.repository.GObject import TYPE_UINT64

    from deluge.ui.gtk3.gtkui import DEFAULT_PREFS
    from deluge.ui.gtk3.mainwindow import MainWindow
    from deluge.ui.gtk3.menubar import MenuBar
    from deluge.ui.gtk3.torrentdetails import TorrentDetails
    from deluge.ui.gtk3.torrentview import TorrentView
except (ImportError, ValueError):
    libs_available = False
    TYPE_UINT64 = 'Whatever'
else:
    libs_available = True

setup_translation()


@pytest.mark.gtkui
class TestTorrentview(BaseTestCase):

    default_column_index = [
        'filter',
        'torrent_id',
        'dirty',
        '#',
        'Name',
        'Size',
        'Downloaded',
        'Uploaded',
        'Remaining',
        'Progress',
        'Seeds',
        'Peers',
        'Seeds:Peers',
        'Down Speed',
        'Up Speed',
        'Down Limit',
        'Up Limit',
        'ETA',
        'Ratio',
        'Avail',
        'Added',
        'Completed',
        'Complete Seen',
        'Last Transfer',
        'Tracker',
        'Download Folder',
        'Owner',
        'Shared',
    ]
    default_liststore_columns = [
        bool,
        str,
        bool,
        int,
        str,
        str,  # Name
        TYPE_UINT64,
        TYPE_UINT64,
        TYPE_UINT64,
        TYPE_UINT64,
        float,
        str,  # Progress
        int,
        int,
        int,
        int,
        float,  # Seeds, Peers
        int,
        int,
        float,
        float,
        int,
        float,
        float,  # ETA, Ratio, Avail
        int,
        int,
        int,
        int,
        str,
        str,  # Tracker
        str,
        str,
        bool,
    ]  # shared

    def set_up(self):
        if libs_available is False:
            pytest.skip('GTKUI dependencies not available')

        # MainWindow loads this config file, so lets make sure it contains the defaults
        ConfigManager('gtk3ui.conf', defaults=DEFAULT_PREFS)
        self.mainwindow = MainWindow()
        self.torrentview = TorrentView()
        self.torrentdetails = TorrentDetails()
        self.menubar = MenuBar()

    def tear_down(self):
        return component.shutdown()

    def test_torrentview_columns(self):
        assert self.torrentview.column_index == self.default_column_index
        assert self.torrentview.liststore_columns == self.default_liststore_columns
        assert self.torrentview.columns['Download Folder'].column_indices == [30]

    def test_add_column(self):
        # Add a text column
        test_col = 'Test column'
        self.torrentview.add_text_column(test_col, status_field=['label'])
        assert (
            len(self.torrentview.liststore_columns)
            == len(self.default_liststore_columns) + 1
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index) + 1
        assert self.torrentview.column_index[-1] == test_col
        assert self.torrentview.columns[test_col].column_indices == [33]

    def test_add_columns(self):
        # Add a text column
        test_col = 'Test column'
        self.torrentview.add_text_column(test_col, status_field=['label'])

        # Add a second text column
        test_col2 = 'Test column2'
        self.torrentview.add_text_column(test_col2, status_field=['label2'])

        assert (
            len(self.torrentview.liststore_columns)
            == len(self.default_liststore_columns) + 2
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index) + 2
        # test_col
        assert self.torrentview.column_index[-2] == test_col
        assert self.torrentview.columns[test_col].column_indices == [33]

        # test_col2
        assert self.torrentview.column_index[-1] == test_col2
        assert self.torrentview.columns[test_col2].column_indices == [34]

    def test_remove_column(self):
        # Add and remove text column
        test_col = 'Test column'
        self.torrentview.add_text_column(test_col, status_field=['label'])
        self.torrentview.remove_column(test_col)

        assert len(self.torrentview.liststore_columns) == len(
            self.default_liststore_columns
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index)
        assert self.torrentview.column_index[-1] == self.default_column_index[-1]
        assert self.torrentview.columns[
            self.default_column_index[-1]
        ].column_indices == [32]

    def test_remove_columns(self):
        # Add two columns
        test_col = 'Test column'
        self.torrentview.add_text_column(test_col, status_field=['label'])
        test_col2 = 'Test column2'
        self.torrentview.add_text_column(test_col2, status_field=['label2'])

        # Remove test_col
        self.torrentview.remove_column(test_col)
        assert (
            len(self.torrentview.liststore_columns)
            == len(self.default_liststore_columns) + 1
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index) + 1
        assert self.torrentview.column_index[-1] == test_col2
        assert self.torrentview.columns[test_col2].column_indices == [33]

        # Remove test_col2
        self.torrentview.remove_column(test_col2)
        assert len(self.torrentview.liststore_columns) == len(
            self.default_liststore_columns
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index)
        assert self.torrentview.column_index[-1] == self.default_column_index[-1]
        assert self.torrentview.columns[
            self.default_column_index[-1]
        ].column_indices == [32]

    def test_add_remove_column_multiple_types(self):
        # Add a column with multiple column types
        test_col3 = 'Test column3'
        self.torrentview.add_progress_column(
            test_col3, status_field=['progress', 'label3'], col_types=[float, str]
        )
        assert (
            len(self.torrentview.liststore_columns)
            == len(self.default_liststore_columns) + 2
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index) + 1
        assert self.torrentview.column_index[-1] == test_col3
        assert self.torrentview.columns[test_col3].column_indices == [33, 34]

        # Remove multiple column-types column
        self.torrentview.remove_column(test_col3)

        assert len(self.torrentview.liststore_columns) == len(
            self.default_liststore_columns
        )
        assert len(self.torrentview.column_index) == len(self.default_column_index)
        assert self.torrentview.column_index[-1] == self.default_column_index[-1]
        assert self.torrentview.columns[
            self.default_column_index[-1]
        ].column_indices == [32]



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from twisted.internet import defer

import deluge.component as component

from . import BaseCommand

log = logging.getLogger(__name__)


class Command(BaseCommand):
    """Displays help on other commands"""

    def add_arguments(self, parser):
        parser.add_argument(
            'commands', metavar='<command>', nargs='*', help=_('One or more commands')
        )

    def handle(self, options):
        self.console = component.get('ConsoleUI')
        self._commands = self.console._commands
        deferred = defer.succeed(True)
        if options.commands:
            for arg in options.commands:
                try:
                    cmd = self._commands[arg]
                except KeyError:
                    self.console.write('{!error!}Unknown command %s' % arg)
                    continue
                try:
                    parser = cmd.create_parser()
                    self.console.write(parser.format_help())
                except AttributeError:
                    self.console.write(cmd.__doc__ or 'No help for this command')
                self.console.write(' ')
        else:
            self.console.set_batch_write(True)
            cmds_doc = ''
            for cmd in sorted(self._commands):
                if cmd in self._commands[cmd].aliases:
                    continue
                parser = self._commands[cmd].create_parser()
                cmd_doc = (
                    '{!info!}'
                    + '%-9s' % self._commands[cmd].name_with_alias
                    + '{!input!} - '
                    + self._commands[cmd].__doc__
                    + '\n     '
                    + parser.format_usage()
                    or ''
                )
                cmds_doc += parser.formatter.format_colors(cmd_doc)
            self.console.write(cmds_doc)
            self.console.write(' ')
            self.console.write('For help on a specific command, use `<command> --help`')
            self.console.set_batch_write(False)

        return deferred

    def complete(self, line):
        return [x for x in component.get('ConsoleUI')._commands if x.startswith(line)]



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge.plugins.pluginbase import WebPluginBase

from .common import get_resource

log = logging.getLogger(__name__)

FORMAT_LIST = [
    ('gzmule', _('Emule IP list (GZip)')),
    ('spzip', _('SafePeer Text (Zipped)')),
    ('pgtext', _('PeerGuardian Text (Uncompressed)')),
    ('p2bgz', _('PeerGuardian P2B (GZip)')),
]


class WebUI(WebPluginBase):

    scripts = [get_resource('blocklist.js')]
    debug_scripts = scripts



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import pytest

import deluge.component as component
from deluge.configmanager import ConfigManager
from deluge.conftest import BaseTestCase
from deluge.i18n import setup_translation

libs_available = True
# Allow running other tests without GTKUI dependencies available
try:
    from deluge.ui.gtk3.files_tab import FilesTab
    from deluge.ui.gtk3.gtkui import DEFAULT_PREFS
    from deluge.ui.gtk3.mainwindow import MainWindow
except (ImportError, ValueError):
    # gi.require_version gives ValueError if library not available
    libs_available = False

setup_translation()


@pytest.mark.gtkui
class TestFilesTab(BaseTestCase):
    def set_up(self):
        if libs_available is False:
            pytest.skip('GTKUI dependencies not available')

        ConfigManager('gtk3ui.conf', defaults=DEFAULT_PREFS)
        self.mainwindow = MainWindow()
        self.filestab = FilesTab()
        self.t_id = '1'
        self.filestab.torrent_id = self.t_id
        self.index = 1

    def tear_down(self):
        return component.shutdown()

    def print_treestore(self, title, treestore):
        root = treestore.get_iter_first()
        level = 1

        def p_level(s, lvl):
            print('{}{}'.format(' ' * lvl, s))

        def _print_treestore_children(i, lvl):
            while i:
                p_level(treestore[i][0], lvl)
                if treestore.iter_children(i):
                    _print_treestore_children(treestore.iter_children(i), lvl + 2)
                i = treestore.iter_next(i)

        print('\n%s' % title)
        _print_treestore_children(root, level)
        print('')

    def verify_treestore(self, treestore, tree):
        def _verify_treestore(itr, tree_values):
            i = 0
            while itr:
                values = tree_values[i]
                if treestore[itr][0] != values[0]:
                    return False
                if treestore.iter_children(itr):
                    if not _verify_treestore(treestore.iter_children(itr), values[1]):
                        return False
                itr = treestore.iter_next(itr)
                i += 1
            return True

        return _verify_treestore(treestore.get_iter_first(), tree)

    def test_files_tab(self):
        self.filestab.files_list[self.t_id] = (
            {'index': 0, 'path': '1/test_10.txt', 'offset': 0, 'size': 13},
            {'index': 1, 'path': 'test_100.txt', 'offset': 13, 'size': 14},
        )
        self.filestab.update_files()
        self.filestab._on_torrentfilerenamed_event(
            self.t_id, self.index, '2/test_100.txt'
        )

        ret = self.verify_treestore(
            self.filestab.treestore,
            [['1/', [['test_10.txt']]], ['2/', [['test_100.txt']]]],
        )
        if not ret:
            self.print_treestore('Treestore not expected:', self.filestab.treestore)
        assert ret

    def test_files_tab2(self):
        self.filestab.files_list[self.t_id] = (
            {'index': 0, 'path': '1/1/test_100.txt', 'offset': 0, 'size': 13},
            {'index': 1, 'path': 'test_101.txt', 'offset': 13, 'size': 14},
        )
        self.filestab.update_files()
        self.filestab._on_torrentfilerenamed_event(
            self.t_id, self.index, '1/1/test_101.txt'
        )

        ret = self.verify_treestore(
            self.filestab.treestore,
            [['1/', [['1/', [['test_100.txt'], ['test_101.txt']]]]]],
        )
        if not ret:
            self.print_treestore('Treestore not expected:', self.filestab.treestore)
        assert ret

    def test_files_tab3(self):
        self.filestab.files_list[self.t_id] = (
            {'index': 0, 'path': '1/test_100.txt', 'offset': 0, 'size': 13},
            {'index': 1, 'path': 'test_101.txt', 'offset': 13, 'size': 14},
        )
        self.filestab.update_files()
        self.filestab._on_torrentfilerenamed_event(
            self.t_id, self.index, '1/test_101.txt'
        )

        ret = self.verify_treestore(
            self.filestab.treestore, [['1/', [['test_100.txt'], ['test_101.txt']]]]
        )
        if not ret:
            self.print_treestore('Treestore not expected:', self.filestab.treestore)
        assert ret

    def test_files_tab4(self):
        self.filestab.files_list[self.t_id] = (
            {'index': 0, 'path': '1/test_100.txt', 'offset': 0, 'size': 13},
            {'index': 1, 'path': '1/test_101.txt', 'offset': 13, 'size': 14},
        )
        self.filestab.update_files()
        self.filestab._on_torrentfilerenamed_event(
            self.t_id, self.index, '1/2/test_101.txt'
        )

        ret = self.verify_treestore(
            self.filestab.treestore,
            [['1/', [['2/', [['test_101.txt']]], ['test_100.txt']]]],
        )
        if not ret:
            self.print_treestore('Treestore not expected:', self.filestab.treestore)
        assert ret

    def test_files_tab5(self):
        self.filestab.files_list[self.t_id] = (
            {'index': 0, 'path': '1/test_100.txt', 'offset': 0, 'size': 13},
            {'index': 1, 'path': '2/test_101.txt', 'offset': 13, 'size': 14},
        )
        self.filestab.update_files()
        self.filestab._on_torrentfilerenamed_event(
            self.t_id, self.index, '1/test_101.txt'
        )

        ret = self.verify_treestore(
            self.filestab.treestore, [['1/', [['test_100.txt'], ['test_101.txt']]]]
        )
        if not ret:
            self.print_treestore('Treestore not expected:', self.filestab.treestore)
        assert ret



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#


from deluge.decorators import proxy


class TestDecorators:
    def test_proxy_with_simple_functions(self):
        def negate(func, *args, **kwargs):
            return not func(*args, **kwargs)

        @proxy(negate)
        def something(_bool):
            return _bool

        @proxy(negate)
        @proxy(negate)
        def double_nothing(_bool):
            return _bool

        assert something(False)
        assert not something(True)
        assert double_nothing(True)
        assert not double_nothing(False)

    def test_proxy_with_class_method(self):
        def negate(func, *args, **kwargs):
            return -func(*args, **kwargs)

        class Test:
            def __init__(self, number):
                self.number = number

            @proxy(negate)
            def diff(self, number):
                return self.number - number

            @proxy(negate)
            def no_diff(self, number):
                return self.diff(number)

        t = Test(5)
        assert t.diff(1) == -4
        assert t.no_diff(1) == 4



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import pytest
import pytest_twisted
from twisted.internet import defer

from deluge import error
from deluge.common import AUTH_LEVEL_NORMAL, get_localhost_auth
from deluge.core.authmanager import AUTH_LEVEL_ADMIN
from deluge.ui.client import Client, DaemonSSLProxy, client


class NoVersionSendingDaemonSSLProxy(DaemonSSLProxy):
    def authenticate(self, username, password):
        self.login_deferred = defer.Deferred()
        d = self.call('daemon.login', username, password)
        d.addCallback(self.__on_login, username)
        d.addErrback(self.__on_login_fail)
        return self.login_deferred

    def __on_login(self, result, username):
        self.login_deferred.callback(result)

    def __on_login_fail(self, result):
        self.login_deferred.errback(result)


class NoVersionSendingClient(Client):
    def connect(
        self,
        host='127.0.0.1',
        port=58846,
        username='',
        password='',
        skip_authentication=False,
    ):
        self._daemon_proxy = NoVersionSendingDaemonSSLProxy()
        self._daemon_proxy.set_disconnect_callback(self.__on_disconnect)

        d = self._daemon_proxy.connect(host, port)

        def on_connect_fail(reason):
            self.disconnect()
            return reason

        def on_authenticate(result, daemon_info):
            return result

        def on_authenticate_fail(reason):
            return reason

        def on_connected(daemon_version):
            return daemon_version

        def authenticate(daemon_version, username, password):
            d = self._daemon_proxy.authenticate(username, password)
            d.addCallback(on_authenticate, daemon_version)
            d.addErrback(on_authenticate_fail)
            return d

        d.addCallback(on_connected)
        d.addErrback(on_connect_fail)
        if not skip_authentication:
            d.addCallback(authenticate, username, password)
        return d

    def __on_disconnect(self):
        if self.disconnect_callback:
            self.disconnect_callback()


@pytest.mark.usefixtures('daemon', 'client')
class TestClient:
    def test_connect_no_credentials(self):
        d = client.connect('localhost', self.listen_port, username='', password='')

        def on_connect(result):
            assert client.get_auth_level() == AUTH_LEVEL_ADMIN
            return result

        d.addCallbacks(on_connect, self.fail)
        return d

    def test_connect_localclient(self):
        username, password = get_localhost_auth()
        d = client.connect(
            'localhost', self.listen_port, username=username, password=password
        )

        def on_connect(result):
            assert client.get_auth_level() == AUTH_LEVEL_ADMIN
            return result

        d.addCallbacks(on_connect, self.fail)
        return d

    def test_connect_bad_password(self):
        username, password = get_localhost_auth()
        d = client.connect(
            'localhost', self.listen_port, username=username, password=password + '1'
        )

        def on_failure(failure):
            assert failure.trap(error.BadLoginError) == error.BadLoginError
            assert failure.value.message == 'Password does not match'

        d.addCallbacks(self.fail, on_failure)
        return d

    def test_connect_invalid_user(self):
        d = client.connect('localhost', self.listen_port, username='invalid-user')

        def on_failure(failure):
            assert failure.trap(error.BadLoginError) == error.BadLoginError
            assert failure.value.message == 'Username does not exist'

        d.addCallbacks(self.fail, on_failure)
        return d

    def test_connect_without_password(self):
        username, password = get_localhost_auth()
        d = client.connect('localhost', self.listen_port, username=username)

        def on_failure(failure):
            assert (
                failure.trap(error.AuthenticationRequired)
                == error.AuthenticationRequired
            )
            assert failure.value.username == username

        d.addCallbacks(self.fail, on_failure)
        return d

    @pytest_twisted.inlineCallbacks
    def test_connect_with_password(self):
        username, password = get_localhost_auth()
        yield client.connect(
            'localhost', self.listen_port, username=username, password=password
        )
        yield client.core.create_account('testuser', 'testpw', 'DEFAULT')
        yield client.disconnect()
        ret = yield client.connect(
            'localhost', self.listen_port, username='testuser', password='testpw'
        )
        assert ret == AUTH_LEVEL_NORMAL

    @pytest_twisted.inlineCallbacks
    def test_invalid_rpc_method_call(self):
        yield client.connect('localhost', self.listen_port, username='', password='')
        d = client.core.invalid_method()

        def on_failure(failure):
            assert failure.trap(error.WrappedException) == error.WrappedException

        d.addCallbacks(self.fail, on_failure)
        yield d

    def test_connect_without_sending_client_version_fails(self):
        username, password = get_localhost_auth()
        no_version_sending_client = NoVersionSendingClient()
        d = no_version_sending_client.connect(
            'localhost', self.listen_port, username=username, password=password
        )

        def on_failure(failure):
            assert failure.trap(error.IncompatibleClient) == error.IncompatibleClient

        d.addCallbacks(self.fail, on_failure)
        return d



#
# Copyright (C) 2010 John Garland <johnnybg+deluge@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.core.rpcserver import export
from deluge.plugins.pluginbase import CorePluginBase

log = logging.getLogger(__name__)

DEFAULT_PREFS = {}


class Core(CorePluginBase):
    def enable(self):
        self.core = component.get('Core')

    def disable(self):
        pass

    def update(self):
        pass

    @export
    def get_status(self):
        return self.core.session.is_paused()

    @export
    def toggle(self):
        if self.core.session.is_paused():
            self.core.resume_session()
            paused = False
        else:
            self.core.pause_session()
            paused = True
        return paused



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import pytest
import pytest_twisted
from twisted.internet import defer, threads

import deluge.component as component


class ComponentTester(component.Component):
    def __init__(self, name, depend=None):
        component.Component.__init__(self, name, depend=depend)
        self.start_count = 0
        self.stop_count = 0

    def start(self):
        self.start_count += 1

    def stop(self):
        self.stop_count += 1


class ComponentTesterDelayStart(ComponentTester):
    def start(self):
        def do_sleep():
            import time

            time.sleep(1)

        d = threads.deferToThread(do_sleep)

        def on_done(result):
            self.start_count += 1

        return d.addCallback(on_done)


class ComponentTesterUpdate(component.Component):
    def __init__(self, name):
        component.Component.__init__(self, name)
        self.counter = 0
        self.start_count = 0
        self.stop_count = 0

    def update(self):
        self.counter += 1

    def stop(self):
        self.stop_count += 1


class ComponentTesterShutdown(component.Component):
    def __init__(self, name):
        component.Component.__init__(self, name)
        self.shutdowned = False
        self.stop_count = 0

    def shutdown(self):
        self.shutdowned = True

    def stop(self):
        self.stop_count += 1


@pytest.mark.usefixtures('component')
class TestComponent:
    def tear_down(self):
        return component.shutdown()

    def test_start_component(self):
        def on_start(result, c):
            assert c._component_state == 'Started'
            assert c.start_count == 1

        c = ComponentTester('test_start_c1')
        d = component.start(['test_start_c1'])
        d.addCallback(on_start, c)
        return d

    def test_start_stop_depends(self):
        def on_stop(result, c1, c2):
            assert c1._component_state == 'Stopped'
            assert c2._component_state == 'Stopped'
            assert c1.stop_count == 1
            assert c2.stop_count == 1

        def on_start(result, c1, c2):
            assert c1._component_state == 'Started'
            assert c2._component_state == 'Started'
            assert c1.start_count == 1
            assert c2.start_count == 1
            return component.stop(['test_start_depends_c1']).addCallback(
                on_stop, c1, c2
            )

        c1 = ComponentTester('test_start_depends_c1')
        c2 = ComponentTester('test_start_depends_c2', depend=['test_start_depends_c1'])

        d = component.start(['test_start_depends_c2'])
        d.addCallback(on_start, c1, c2)
        return d

    def start_with_depends(self):
        c1 = ComponentTesterDelayStart('test_start_all_c1')
        c2 = ComponentTester('test_start_all_c2', depend=['test_start_all_c4'])
        c3 = ComponentTesterDelayStart(
            'test_start_all_c3', depend=['test_start_all_c5', 'test_start_all_c1']
        )
        c4 = ComponentTester('test_start_all_c4', depend=['test_start_all_c3'])
        c5 = ComponentTester('test_start_all_c5')

        d = component.start()
        return (d, c1, c2, c3, c4, c5)

    def finish_start_with_depends(self, *args):
        for c in args[1:]:
            component.deregister(c)

    def test_start_all(self):
        def on_start(*args):
            for c in args[1:]:
                assert c._component_state == 'Started'
                assert c.start_count == 1

        ret = self.start_with_depends()
        ret[0].addCallback(on_start, *ret[1:])
        ret[0].addCallback(self.finish_start_with_depends, *ret[1:])
        return ret[0]

    def test_register_exception(self):
        ComponentTester('test_register_exception_c1')
        with pytest.raises(component.ComponentAlreadyRegistered):
            ComponentTester(
                'test_register_exception_c1',
            )

    def test_stop_component(self):
        def on_stop(result, c):
            assert c._component_state == 'Stopped'
            assert not c._component_timer.running
            assert c.stop_count == 1

        def on_start(result, c):
            assert c._component_state == 'Started'
            return component.stop(['test_stop_component_c1']).addCallback(on_stop, c)

        c = ComponentTesterUpdate('test_stop_component_c1')
        d = component.start(['test_stop_component_c1'])
        d.addCallback(on_start, c)
        return d

    def test_stop_all(self):
        def on_stop(result, *args):
            for c in args:
                assert c._component_state == 'Stopped'
                assert c.stop_count == 1

        def on_start(result, *args):
            for c in args:
                assert c._component_state == 'Started'
            return component.stop().addCallback(on_stop, *args)

        ret = self.start_with_depends()
        ret[0].addCallback(on_start, *ret[1:])
        ret[0].addCallback(self.finish_start_with_depends, *ret[1:])
        return ret[0]

    def test_update(self):
        def on_start(result, c1, counter):
            assert c1._component_timer
            assert c1._component_timer.running
            assert c1.counter != counter
            return component.stop()

        c1 = ComponentTesterUpdate('test_update_c1')
        cnt = int(c1.counter)
        d = component.start(['test_update_c1'])

        d.addCallback(on_start, c1, cnt)
        return d

    def test_pause(self):
        def on_pause(result, c1, counter):
            assert c1._component_state == 'Paused'
            assert c1.counter != counter
            assert not c1._component_timer.running

        def on_start(result, c1, counter):
            assert c1._component_timer
            assert c1.counter != counter
            d = component.pause(['test_pause_c1'])
            d.addCallback(on_pause, c1, counter)
            return d

        c1 = ComponentTesterUpdate('test_pause_c1')
        cnt = int(c1.counter)
        d = component.start(['test_pause_c1'])

        d.addCallback(on_start, c1, cnt)
        return d

    @pytest_twisted.inlineCallbacks
    def test_component_start_error(self):
        ComponentTesterUpdate('test_pause_c1')
        yield component.start(['test_pause_c1'])
        yield component.pause(['test_pause_c1'])
        test_comp = component.get('test_pause_c1')
        with pytest.raises(component.ComponentException, match='Current state: Paused'):
            yield test_comp._component_start()

    @pytest_twisted.inlineCallbacks
    def test_start_paused_error(self):
        ComponentTesterUpdate('test_pause_c1')
        yield component.start(['test_pause_c1'])
        yield component.pause(['test_pause_c1'])

        # Deferreds that fail in component have to error handler which results in
        # twisted doing a log.err call which causes the test to fail.
        # Prevent failure by ignoring the exception
        # self._observer._ignoreErrors(component.ComponentException)

        result = yield component.start()
        assert [(result[0][0], result[0][1].value)] == [
            (
                defer.FAILURE,
                component.ComponentException(
                    'Trying to start component "%s" but it is '
                    'not in a stopped state. Current state: %s'
                    % ('test_pause_c1', 'Paused'),
                    '',
                ),
            )
        ]

    def test_shutdown(self):
        def on_shutdown(result, c1):
            assert c1.shutdowned
            assert c1._component_state == 'Stopped'
            assert c1.stop_count == 1

        def on_start(result, c1):
            d = component.shutdown()
            d.addCallback(on_shutdown, c1)
            return d

        c1 = ComponentTesterShutdown('test_shutdown_c1')
        d = component.start(['test_shutdown_c1'])
        d.addCallback(on_start, c1)
        return d



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component
from deluge.decorators import overrides
from deluge.ui.console.modes.basemode import InputKeyHandler
from deluge.ui.console.modes.torrentlist import torrentviewcolumns
from deluge.ui.console.modes.torrentlist.torrentactions import torrent_actions_popup
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.utils import format_utils
from deluge.ui.console.utils.column import (
    get_column_value,
    get_required_fields,
    torrent_data_fields,
)

from . import ACTION

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)


state_fg_colors = {
    'Downloading': 'green',
    'Seeding': 'cyan',
    'Error': 'red',
    'Queued': 'yellow',
    'Checking': 'blue',
    'Moving': 'green',
}


reverse_sort_fields = [
    'size',
    'download_speed',
    'upload_speed',
    'num_seeds',
    'num_peers',
    'distributed_copies',
    'time_added',
    'total_uploaded',
    'all_time_download',
    'total_remaining',
    'progress',
    'ratio',
    'seeding_time',
    'active_time',
]


default_column_values = {
    'queue': {'width': 4, 'visible': True},
    'name': {'width': -1, 'visible': True},
    'size': {'width': 8, 'visible': True},
    'progress': {'width': 7, 'visible': True},
    'download_speed': {'width': 7, 'visible': True},
    'upload_speed': {'width': 7, 'visible': True},
    'state': {'width': 13},
    'eta': {'width': 8, 'visible': True},
    'time_added': {'width': 15},
    'tracker': {'width': 15},
    'download_location': {'width': 15},
    'downloaded': {'width': 13},
    'uploaded': {'width': 7},
    'remaining': {'width': 13},
    'completed_time': {'width': 15},
    'last_seen_complete': {'width': 15},
    'max_upload_speed': {'width': 7},
}


default_columns = {}
for col_i, col_name in enumerate(torrentviewcolumns.column_pref_names):
    default_columns[col_name] = {'width': 10, 'order': col_i, 'visible': False}
    if col_name in default_column_values:
        default_columns[col_name].update(default_column_values[col_name])


class TorrentView(InputKeyHandler):
    def __init__(self, torrentlist, config):
        super().__init__()
        self.torrentlist = torrentlist
        self.config = config
        self.filter_dict = {}
        self.curr_filter = None
        self.cached_rows = {}
        self.sorted_ids = None
        self.torrent_names = None
        self.numtorrents = -1
        self.column_string = ''
        self.curoff = 0
        self.marked = []
        self.cursel = 0

    @property
    def rows(self):
        return self.torrentlist.rows

    @property
    def torrent_rows(self):
        return self.torrentlist.rows - 3  # Account for header lines + columns line

    @property
    def torrentlist_offset(self):
        return 2

    def update_state(self, state, refresh=False):
        self.curstate = state  # cache in case we change sort order
        self.cached_rows.clear()
        self.numtorrents = len(state)
        self.sorted_ids = self._sort_torrents(state)
        self.torrent_names = []
        for torrent_id in self.sorted_ids:
            ts = self.curstate[torrent_id]
            self.torrent_names.append(ts['name'])

        if refresh:
            self.torrentlist.refresh()

    def set_torrent_filter(self, state):
        self.curr_filter = state
        filter_dict = {'state': [state]}
        if state == 'All':
            self.curr_filter = None
            filter_dict = {}
        self.filter_dict = filter_dict
        self.torrentlist.go_top = True
        self.torrentlist.update()
        return True

    def _scroll_up(self, by):
        cursel = self.cursel
        prevoff = self.curoff
        self.cursel = max(self.cursel - by, 0)
        if self.cursel < self.curoff:
            self.curoff = self.cursel
        affected = []
        if prevoff == self.curoff:
            affected.append(cursel)
            if cursel != self.cursel:
                affected.insert(0, self.cursel)
        return affected

    def _scroll_down(self, by):
        cursel = self.cursel
        prevoff = self.curoff
        self.cursel = min(self.cursel + by, self.numtorrents - 1)
        if (self.curoff + self.torrent_rows) <= self.cursel:
            self.curoff = self.cursel - self.torrent_rows + 1
        affected = []
        if prevoff == self.curoff:
            affected.append(cursel)
            if cursel != self.cursel:
                affected.append(self.cursel)
        return affected

    def current_torrent_id(self):
        if not self.sorted_ids:
            return None
        return self.sorted_ids[self.cursel]

    def _selected_torrent_ids(self):
        if not self.sorted_ids:
            return None
        ret = []
        for i in self.marked:
            ret.append(self.sorted_ids[i])
        return ret

    def clear_marked(self):
        self.marked = []
        self.last_mark = -1

    def mark_unmark(self, idx):
        if idx in self.marked:
            self.marked.remove(idx)
            self.last_mark = -1
        else:
            self.marked.append(idx)
            self.last_mark = idx

    def add_marked(self, indices, last_marked):
        for i in indices:
            if i not in self.marked:
                self.marked.append(i)
        self.last_mark = last_marked

    def update_marked(self, index, last_mark=True, clear=False):
        if index not in self.marked:
            if clear:
                self.marked = []
            self.marked.append(index)
            if last_mark:
                self.last_mark = index
            return True
        return False

    def _sort_torrents(self, state):
        """Sorts by primary and secondary sort fields."""

        if not state:
            return {}

        s_primary = self.config['torrentview']['sort_primary']
        s_secondary = self.config['torrentview']['sort_secondary']

        result = state

        # Sort first by secondary sort field and then primary sort field
        # so it all works out

        def sort_by_field(state, to_sort, field):
            field = torrent_data_fields[field]['status'][0]
            reverse = field in reverse_sort_fields

            # Get first element so we can check if it has given field
            # and if it's a string
            first_element = state[list(state)[0]]
            if field in first_element:

                def sort_key(s):
                    try:
                        # Sort case-insensitively but preserve A>a order.
                        return state.get(s)[field].lower()
                    except AttributeError:
                        # Not a string.
                        return state.get(s)[field]

                to_sort = sorted(to_sort, key=sort_key, reverse=reverse)

            if field == 'eta':
                to_sort = sorted(to_sort, key=lambda s: state.get(s)['eta'] == 0)

            return to_sort

        # Just in case primary and secondary fields are empty and/or
        # both are too ambiguous, also sort by queue position first
        if 'queue' not in [s_secondary, s_primary]:
            result = sort_by_field(state, result, 'queue')
        if s_secondary != s_primary:
            result = sort_by_field(state, result, s_secondary)
        result = sort_by_field(state, result, s_primary)

        if self.config['torrentview']['separate_complete']:
            result = sorted(
                result, key=lambda s: state.get(s).get('progress', 0) == 100.0
            )

        return result

    def _get_colors(self, row, tidx):
        # default style
        colors = {'fg': 'white', 'bg': 'black', 'attr': None}

        if tidx in self.marked:
            colors.update({'bg': 'blue', 'attr': 'bold'})

        if tidx == self.cursel:
            col_selected = {'bg': 'white', 'fg': 'black', 'attr': 'bold'}
            if tidx in self.marked:
                col_selected['fg'] = 'blue'
            colors.update(col_selected)

        colors['fg'] = state_fg_colors.get(row[1], colors['fg'])

        if self.torrentlist.minor_mode:
            self.torrentlist.minor_mode.update_colors(tidx, colors)
        return colors

    def update_torrents(self, lines):
        # add all the torrents
        if self.numtorrents == 0:
            cols = self.torrentlist.torrentview_columns()
            msg = 'No torrents match filter'.center(cols)
            self.torrentlist.add_string(
                3, '{!info!}%s' % msg, scr=self.torrentlist.torrentview_panel
            )
        elif self.numtorrents == 0:
            self.torrentlist.add_string(1, 'Waiting for torrents from core...')
            return

        def draw_row(index):
            if index not in self.cached_rows:
                ts = self.curstate[self.sorted_ids[index]]
                self.cached_rows[index] = (
                    format_utils.format_row(
                        [get_column_value(name, ts) for name in self.cols_to_show],
                        self.column_widths,
                    ),
                    ts['state'],
                )
            return self.cached_rows[index]

        tidx = self.curoff
        currow = 0
        todraw = []
        # Affected lines are given when changing selected torrent
        if lines:
            for line in lines:
                if line < tidx:
                    continue
                if line >= (tidx + self.torrent_rows) or line >= self.numtorrents:
                    break
                todraw.append((line, line - self.curoff, draw_row(line)))
        else:
            for i in range(tidx, tidx + self.torrent_rows):
                if i >= self.numtorrents:
                    break
                todraw.append((i, i - self.curoff, draw_row(i)))

        for tidx, currow, row in todraw:
            if (currow + self.torrentlist_offset - 1) > self.torrent_rows:
                continue
            colors = self._get_colors(row, tidx)
            if colors['attr']:
                colorstr = '{!%(fg)s,%(bg)s,%(attr)s!}' % colors
            else:
                colorstr = '{!%(fg)s,%(bg)s!}' % colors

            self.torrentlist.add_string(
                currow + self.torrentlist_offset,
                f'{colorstr}{row[0]}',
                trim=False,
                scr=self.torrentlist.torrentview_panel,
            )

    def update(self, refresh=False):
        d = component.get('SessionProxy').get_torrents_status(
            self.filter_dict, self.status_fields
        )
        d.addCallback(self.update_state, refresh=refresh)

    def on_config_changed(self):
        s_primary = self.config['torrentview']['sort_primary']
        s_secondary = self.config['torrentview']['sort_secondary']
        changed = None
        for col in default_columns:
            if col not in self.config['torrentview']['columns']:
                changed = self.config['torrentview']['columns'][col] = default_columns[
                    col
                ]
        if changed:
            self.config.save()

        self.cols_to_show = [
            col
            for col in sorted(
                self.config['torrentview']['columns'],
                key=lambda k: self.config['torrentview']['columns'][k]['order'],
            )
            if self.config['torrentview']['columns'][col]['visible']
        ]
        self.status_fields = get_required_fields(self.cols_to_show)

        # we always need these, even if we're not displaying them
        for rf in ['state', 'name', 'queue', 'progress']:
            if rf not in self.status_fields:
                self.status_fields.append(rf)

        # same with sort keys
        if s_primary and s_primary not in self.status_fields:
            self.status_fields.append(s_primary)
        if s_secondary and s_secondary not in self.status_fields:
            self.status_fields.append(s_secondary)

        self.update_columns()

    def update_columns(self):
        self.column_widths = [
            self.config['torrentview']['columns'][col]['width']
            for col in self.cols_to_show
        ]
        requested_width = sum(width for width in self.column_widths if width >= 0)

        cols = self.torrentlist.torrentview_columns()
        if requested_width > cols:  # can't satisfy requests, just spread out evenly
            cw = int(cols / len(self.cols_to_show))
            for i in range(0, len(self.column_widths)):
                self.column_widths[i] = cw
        else:
            rem = cols - requested_width
            var_cols = len([width for width in self.column_widths if width < 0])
            if var_cols > 0:
                vw = int(rem / var_cols)
                for i in range(0, len(self.column_widths)):
                    if self.column_widths[i] < 0:
                        self.column_widths[i] = vw

        self.column_string = '{!header!}'

        primary_sort_col_name = self.config['torrentview']['sort_primary']

        for i, column in enumerate(self.cols_to_show):
            ccol = torrent_data_fields[column]['name']
            width = self.column_widths[i]

            # Trim the column if it's too long to fit
            if len(ccol) > width:
                ccol = ccol[: width - 1]

            # Padding
            ccol += ' ' * (width - len(ccol))

            # Highlight the primary sort column
            if column == primary_sort_col_name:
                if i != len(self.cols_to_show) - 1:
                    ccol = '{!black,green,bold!}%s{!header!}' % ccol
                else:
                    ccol = ('{!black,green,bold!}%s' % ccol)[:-1]

            self.column_string += ccol

    @overrides(InputKeyHandler)
    def handle_read(self, c):
        affected_lines = None
        if c == curses.KEY_UP:
            if self.cursel != 0:
                affected_lines = self._scroll_up(1)
        elif c == curses.KEY_PPAGE:
            affected_lines = self._scroll_up(int(self.torrent_rows / 2))
        elif c == curses.KEY_DOWN:
            if self.cursel < self.numtorrents:
                affected_lines = self._scroll_down(1)
        elif c == curses.KEY_NPAGE:
            affected_lines = self._scroll_down(int(self.torrent_rows / 2))
        elif c == curses.KEY_HOME:
            affected_lines = self._scroll_up(self.cursel)
        elif c == curses.KEY_END:
            affected_lines = self._scroll_down(self.numtorrents - self.cursel)
        elif c == curses.KEY_DC:  # DEL
            added = self.update_marked(self.cursel)

            def on_close(**kwargs):
                if added:
                    self.marked.pop()

            torrent_actions_popup(
                self.torrentlist,
                self._selected_torrent_ids(),
                action=ACTION.REMOVE,
                close_cb=on_close,
            )
        elif c in [curses.KEY_ENTER, util.KEY_ENTER2] and self.numtorrents:
            added = self.update_marked(self.cursel)

            def on_close(data, **kwargs):
                if added:
                    self.marked.remove(self.cursel)

            torrent_actions_popup(
                self.torrentlist,
                self._selected_torrent_ids(),
                details=True,
                close_cb=on_close,
            )
            self.torrentlist.refresh()
        elif c == ord('j'):
            affected_lines = self._scroll_down(1)
        elif c == ord('k'):
            affected_lines = self._scroll_up(1)
        elif c == ord('m'):
            self.mark_unmark(self.cursel)
            affected_lines = [self.cursel]
        elif c == ord('M'):
            if self.last_mark >= 0:
                if self.cursel > self.last_mark:
                    mrange = list(range(self.last_mark, self.cursel + 1))
                else:
                    mrange = list(range(self.cursel, self.last_mark))
                self.add_marked(mrange, self.cursel)
                affected_lines = mrange
            else:
                self.mark_unmark(self.cursel)
                affected_lines = [self.cursel]
        elif c == ord('c'):
            self.clear_marked()
        elif c == ord('o'):
            if not self.marked:
                added = self.update_marked(self.cursel, clear=True)
            else:
                self.last_mark = -1
            torrent_actions_popup(
                self.torrentlist,
                self._selected_torrent_ids(),
                action=ACTION.TORRENT_OPTIONS,
            )
        elif c in [ord('>'), ord('<')]:
            try:
                i = self.cols_to_show.index(self.config['torrentview']['sort_primary'])
            except ValueError:
                i = 0 if chr(c) == '<' else len(self.cols_to_show)
            else:
                i += 1 if chr(c) == '>' else -1

            i = max(0, min(len(self.cols_to_show) - 1, i))
            self.config['torrentview']['sort_primary'] = self.cols_to_show[i]
            self.config.save()
            self.on_config_changed()
            self.update_columns()
            self.torrentlist.refresh([])
        else:
            return util.ReadState.IGNORED

        self.set_input_result(affected_lines)
        return util.ReadState.CHANGED if affected_lines else util.ReadState.READ



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.ui.console.console import Console

UI_PATH = __path__[0]


def start():

    return Console().start()



# Increase open file descriptor limit to allow tests to run
# without getting error: what():  epoll: Too many open files
from deluge.i18n import setup_translation

try:
    import resource
except ImportError:  # Does not exist on Windows
    pass
else:
    try:
        resource.setrlimit(resource.RLIMIT_NOFILE, (65536, 65536))
    except (ValueError, resource.error) as ex:
        error = 'Failed to raise file descriptor limit: %s' % ex
        # print(error)

# Initialize gettext
setup_translation()



#
# Copyright (C) 2009 GazpachoKing <chase.sterling@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from deluge.plugins.pluginbase import WebPluginBase

from .common import get_resource

log = logging.getLogger(__name__)


class WebUI(WebPluginBase):
    scripts = [
        get_resource('autoadd.js'),
        get_resource('autoadd_options.js'),
        get_resource('main_tab.js', True),
        get_resource('options_tab.js', True),
    ]

    def enable(self):
        pass

    def disable(self):
        pass



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import itertools
import os
import time
from base64 import b64encode
from unittest import mock

import pytest
import pytest_twisted
from twisted.internet import defer, reactor
from twisted.internet.task import deferLater

import deluge.component as component
import deluge.core.torrent
import deluge.tests.common as common
from deluge._libtorrent import lt
from deluge.common import VersionSplit, utf8_encode_structure
from deluge.conftest import BaseTestCase
from deluge.core.core import Core
from deluge.core.rpcserver import RPCServer
from deluge.core.torrent import Torrent
from deluge.core.torrentmanager import TorrentManager, TorrentState


class TestTorrent(BaseTestCase):
    def setup_config(self):
        core_config = deluge.config.Config(
            'core.conf',
            defaults=deluge.core.preferencesmanager.DEFAULT_PREFS,
            config_dir=self.config_dir,
        )
        core_config.save()

    def set_up(self):
        self.setup_config()
        self.rpcserver = RPCServer(listen=False)
        self.core = Core()
        self.core.config.config['lsd'] = False
        self.core.config.config['new_release_check'] = False
        self.session = self.core.session
        self.torrent = None
        return component.start()

    def tear_down(self):
        def on_shutdown(result):
            del self.rpcserver
            del self.core

        return component.shutdown().addCallback(on_shutdown)

    def print_priority_list(self, priorities):
        tmp = ''
        for i, p in enumerate(priorities):
            if i % 100 == 0:
                print(tmp)
                tmp = ''
            tmp += '%s' % p
        print(tmp)

    def assert_state(self, torrent, state):
        torrent.update_state()
        assert torrent.state == state

    def get_torrent_atp(self, filename):
        filename = common.get_test_data_file(filename)
        with open(filename, 'rb') as _file:
            info = lt.torrent_info(lt.bdecode(_file.read()))
        atp = {
            'ti': info,
            'save_path': os.getcwd(),
            'storage_mode': lt.storage_mode_t.storage_mode_sparse,
            'flags': (
                lt.add_torrent_params_flags_t.flag_auto_managed
                | lt.add_torrent_params_flags_t.flag_duplicate_is_error
                & ~lt.add_torrent_params_flags_t.flag_paused
            ),
        }
        return atp

    @pytest_twisted.ensureDeferred
    async def test_set_file_priorities(self):
        if getattr(lt, 'file_prio_alert', None):
            # Libtorrent 2.0.3 and later has a file_prio_alert
            prios_set = defer.Deferred()
            prios_set.addTimeout(1.5, reactor)
            component.get('AlertManager').register_handler(
                'file_prio_alert', lambda a: prios_set.callback(True)
            )
        else:
            # On older libtorrent, we just wait a while
            prios_set = deferLater(reactor, 0.8)

        atp = self.get_torrent_atp('dir_with_6_files.torrent')
        handle = self.session.add_torrent(atp)
        torrent = Torrent(handle, {})

        result = torrent.get_file_priorities()
        assert all(x == 4 for x in result)

        new_priorities = [3, 1, 2, 0, 5, 6, 7]
        torrent.set_file_priorities(new_priorities)
        assert torrent.get_file_priorities() == new_priorities

        # Test with handle.piece_priorities as handle.file_priorities async
        # updates and will return old value. Also need to remove a priority
        # value as one file is much smaller than piece size so doesn't show.
        await prios_set  # Delay to wait for alert from lt
        piece_prio = handle.get_piece_priorities()
        result = all(p in piece_prio for p in [3, 2, 0, 5, 6, 7])
        assert result

    def test_set_prioritize_first_last_pieces(self):
        piece_indexes = [
            0,
            1,
            50,
            51,
            52,
            110,
            111,
            112,
            113,
            200,
            201,
            202,
            212,
            213,
            214,
            215,
            216,
            217,
            457,
            458,
            459,
            460,
            461,
            462,
        ]
        self.run_test_set_prioritize_first_last_pieces(
            'dir_with_6_files.torrent', piece_indexes
        )

    def run_test_set_prioritize_first_last_pieces(
        self, torrent_file, prioritized_piece_indexes
    ):
        atp = self.get_torrent_atp(torrent_file)
        handle = self.session.add_torrent(atp)

        self.torrent = Torrent(handle, {})
        priorities_original = handle.get_piece_priorities()
        self.torrent.set_prioritize_first_last_pieces(True)
        priorities = handle.get_piece_priorities()

        # The length of the list of new priorites is the same as the original
        assert len(priorities_original) == len(priorities)

        # Test the priority of all the pieces against the calculated indexes.
        for idx, priority in enumerate(priorities):
            if idx in prioritized_piece_indexes:
                assert priorities[idx] == 7
            else:
                assert priorities[idx] == 4

        # self.print_priority_list(priorities)

    def test_set_prioritize_first_last_pieces_false(self):
        atp = self.get_torrent_atp('dir_with_6_files.torrent')
        handle = self.session.add_torrent(atp)
        self.torrent = Torrent(handle, {})
        # First set some pieces prioritized
        self.torrent.set_prioritize_first_last_pieces(True)
        # Reset pirorities
        self.torrent.set_prioritize_first_last_pieces(False)
        priorities = handle.get_piece_priorities()

        # Test the priority of the prioritized pieces
        for i in priorities:
            assert priorities[i] == 4

        # self.print_priority_list(priorities)

    def test_torrent_error_data_missing(self):
        options = {'seed_mode': True}
        filename = common.get_test_data_file('test_torrent.file.torrent')
        with open(filename, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id = self.core.add_torrent_file(filename, filedump, options)
        torrent = self.core.torrentmanager.torrents[torrent_id]

        # time.sleep(0.5)  # Delay to wait for lt to finish check on Travis.
        # self.assert_state(torrent, 'Seeding')

        # Force an error by reading (non-existant) piece from disk
        torrent.handle.read_piece(0)
        time.sleep(0.2)  # Delay to wait for alert from lt
        self.assert_state(torrent, 'Error')

    def test_torrent_error_resume_original_state(self):
        options = {'seed_mode': True, 'add_paused': True}
        filename = common.get_test_data_file('test_torrent.file.torrent')
        with open(filename, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id = self.core.add_torrent_file(filename, filedump, options)
        torrent = self.core.torrentmanager.torrents[torrent_id]

        orig_state = 'Paused'
        self.assert_state(torrent, orig_state)

        # Force an error by reading (non-existant) piece from disk
        torrent.handle.read_piece(0)
        time.sleep(0.2)  # Delay to wait for alert from lt
        self.assert_state(torrent, 'Error')

        # Clear error and verify returned to original state
        torrent.force_recheck()

    def test_torrent_error_resume_data_unaltered(self):
        if VersionSplit(lt.__version__) >= VersionSplit('1.2.0.0'):
            pytest.skip('Test not working as expected on lt 1.2 or greater')

        resume_data = {
            'active_time': 13399,
            'num_incomplete': 16777215,
            'announce_to_lsd': 1,
            'seed_mode': 0,
            'pieces': '\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01',
            'paused': 0,
            'seeding_time': 13399,
            'last_scrape': 13399,
            'info-hash': '-\xc5\xd0\xe7\x1af\xfeid\x9ad\r9\xcb\x00\xa2YpIs',
            'max_uploads': 16777215,
            'max_connections': 16777215,
            'num_downloaders': 16777215,
            'total_downloaded': 0,
            'file-format': 'libtorrent resume file',
            'peers6': '',
            'added_time': 1411826665,
            'banned_peers6': '',
            'file_priority': [1],
            'last_seen_complete': 0,
            'total_uploaded': 0,
            'piece_priority': '\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01',
            'file-version': 1,
            'announce_to_dht': 1,
            'auto_managed': 1,
            'upload_rate_limit': 0,
            'completed_time': 1411826665,
            'allocation': 'sparse',
            'blocks per piece': 2,
            'download_rate_limit': 0,
            'libtorrent-version': '0.16.17.0',
            'banned_peers': '',
            'num_seeds': 16777215,
            'sequential_download': 0,
            'announce_to_trackers': 1,
            'peers': '\n\x00\x02\x0f=\xc6SC\x17]\xd8}\x7f\x00\x00\x01=\xc6',
            'finished_time': 13399,
            'last_upload': 13399,
            'trackers': [[]],
            'super_seeding': 0,
            'file sizes': [[512000, 1411826586]],
            'last_download': 13399,
        }
        torrent_state = TorrentState(
            torrent_id='2dc5d0e71a66fe69649a640d39cb00a259704973',
            filename='test_torrent.file.torrent',
            name='',
            save_path='/home/ubuntu/Downloads',
            file_priorities=[1],
            is_finished=True,
        )

        filename = common.get_test_data_file('test_torrent.file.torrent')
        with open(filename, 'rb') as _file:
            filedump = _file.read()
        resume_data = utf8_encode_structure(resume_data)
        torrent_id = self.core.torrentmanager.add(
            state=torrent_state, filedump=filedump, resume_data=lt.bencode(resume_data)
        )
        torrent = self.core.torrentmanager.torrents[torrent_id]

        def assert_resume_data():
            self.assert_state(torrent, 'Error')
            tm_resume_data = lt.bdecode(
                self.core.torrentmanager.resume_data[torrent.torrent_id]
            )
            assert tm_resume_data == resume_data

        return deferLater(reactor, 0.5, assert_resume_data)

    def test_get_eta_seeding(self):
        atp = self.get_torrent_atp('test_torrent.file.torrent')
        handle = self.session.add_torrent(atp)
        self.torrent = Torrent(handle, {})
        assert self.torrent.get_eta() == 0
        self.torrent.status = mock.MagicMock()

        self.torrent.status.upload_payload_rate = 5000
        self.torrent.status.download_payload_rate = 0
        self.torrent.status.all_time_download = 10000
        self.torrent.status.all_time_upload = 500
        self.torrent.is_finished = True
        self.torrent.options = {'stop_at_ratio': False}
        # Test finished and uploading but no stop_at_ratio set.
        assert self.torrent.get_eta() == 0

        self.torrent.options = {'stop_at_ratio': True, 'stop_ratio': 1.5}
        result = self.torrent.get_eta()
        assert result == 2
        assert isinstance(result, int)

    def test_get_eta_downloading(self):
        atp = self.get_torrent_atp('test_torrent.file.torrent')
        handle = self.session.add_torrent(atp)
        self.torrent = Torrent(handle, {})
        assert self.torrent.get_eta() == 0

        self.torrent.status = mock.MagicMock()
        self.torrent.status.download_payload_rate = 50
        self.torrent.status.total_wanted = 10000
        self.torrent.status.total_wanted_done = 5000

        result = self.torrent.get_eta()
        assert result == 100
        assert isinstance(result, int)

    def test_get_name_unicode(self):
        """Test retrieving a unicode torrent name from libtorrent."""
        atp = self.get_torrent_atp('unicode_file.torrent')
        handle = self.session.add_torrent(atp)
        self.torrent = Torrent(handle, {})
        assert self.torrent.get_name() == 'সুকুমার রায়.txt'

    def test_rename_unicode(self):
        """Test renaming file/folders with unicode filenames."""
        atp = self.get_torrent_atp('unicode_filenames.torrent')
        handle = self.session.add_torrent(atp)
        self.torrent = Torrent(handle, {})
        # Ignore TorrentManager method call
        TorrentManager.save_resume_data = mock.MagicMock

        result = self.torrent.rename_folder('unicode_filenames', 'Горбачёв')
        assert isinstance(result, defer.DeferredList)

        result = self.torrent.rename_files([[0, 'new_рбачёв']])
        assert result is None

    def test_connect_peer_port(self):
        """Test to ensure port is int for libtorrent"""
        atp = self.get_torrent_atp('test_torrent.file.torrent')
        handle = self.session.add_torrent(atp)
        self.torrent = Torrent(handle, {})
        assert not self.torrent.connect_peer('127.0.0.1', 'text')
        assert self.torrent.connect_peer('127.0.0.1', '1234')

    def test_status_cache(self):
        atp = self.get_torrent_atp('test_torrent.file.torrent')
        handle = self.session.add_torrent(atp)
        mock_time = mock.Mock(return_value=time.time())
        with mock.patch('time.time', mock_time):
            torrent = Torrent(handle, {})
            counter = itertools.count()
            handle.status = mock.Mock(side_effect=counter.__next__)
            first_status = torrent.get_lt_status()
            assert first_status == 0, 'sanity check'
            assert first_status == torrent.status, 'cached status should be used'
            assert torrent.get_lt_status() == 1, 'status should update'
            assert torrent.status == 1
            # Advance time and verify cache expires and updates
            mock_time.return_value += 10
            assert torrent.status == 2



#
# Copyright (C) 2007 Steve 'Tarka' Smith (tarka@internode.on.net)
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import gzip
import logging
import socket
from struct import unpack

log = logging.getLogger(__name__)


class PGException(Exception):
    pass


# Incrementally reads PeerGuardian blocklists v1 and v2.
# See http://wiki.phoenixlabs.org/wiki/P2B_Format
class PGReader:
    def __init__(self, filename):
        log.debug('PGReader loading: %s', filename)

        try:
            with gzip.open(filename, 'rb') as _file:
                self.fd = _file
        except OSError:
            log.debug('Blocklist: PGReader: Incorrect file type or list is corrupt')

        # 4 bytes, should be 0xffffffff
        buf = self.fd.read(4)
        hdr = unpack('l', buf)[0]
        if hdr != -1:
            raise PGException(_('Invalid leader') + ' %d' % hdr)

        magic = self.fd.read(3)
        if magic != 'P2B':
            raise PGException(_('Invalid magic code'))

        buf = self.fd.read(1)
        ver = ord(buf)
        if ver != 1 and ver != 2:
            raise PGException(_('Invalid version') + ' %d' % ver)

    def __next__(self):
        # Skip over the string
        buf = -1
        while buf != 0:
            buf = self.fd.read(1)
            if buf == '':  # EOF
                return False
            buf = ord(buf)

        buf = self.fd.read(4)
        start = socket.inet_ntoa(buf)

        buf = self.fd.read(4)
        end = socket.inet_ntoa(buf)

        return (start, end)

    def close(self):
        self.fd.close()



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository.Gtk import SeparatorToolItem, ToolButton

import deluge.component as component
from deluge.configmanager import ConfigManager

log = logging.getLogger(__name__)


class ToolBar(component.Component):
    def __init__(self):
        component.Component.__init__(self, 'ToolBar')
        log.debug('ToolBar Init..')
        mainwindow = component.get('MainWindow')
        self.main_builder = mainwindow.get_builder()
        self.toolbar = self.main_builder.get_object('toolbar')
        self.config = ConfigManager('gtk3ui.conf')
        # Connect main window Signals #
        mainwindow.connect_signals(self)
        self.change_sensitivity = [
            'toolbutton_add',
            'toolbutton_remove',
            'toolbutton_pause',
            'toolbutton_resume',
            'toolbutton_queue_up',
            'toolbutton_queue_down',
            'toolbutton_filter',
            'find_menuitem',
        ]

        # Hide if necessary
        self.visible(self.config['show_toolbar'])

    def start(self):
        self.main_builder.get_object('toolbutton_connectionmanager').set_visible(
            not self.config['standalone']
        )

        for widget in self.change_sensitivity:
            self.main_builder.get_object(widget).set_sensitive(True)

    def stop(self):
        for widget in self.change_sensitivity:
            self.main_builder.get_object(widget).set_sensitive(False)

    def visible(self, visible):
        if visible:
            self.toolbar.show()
        else:
            self.toolbar.hide()

        self.config['show_toolbar'] = visible

    def add_toolbutton(
        self, callback, label=None, image=None, stock=None, tooltip=None
    ):
        """Adds a toolbutton to the toolbar"""
        toolbutton = ToolButton()
        if stock is not None:
            toolbutton.set_stock_id(stock)
        if label is not None:
            toolbutton.set_label(label)
        if image is not None:
            toolbutton.set_icon_widget(image)
        if tooltip is not None:
            toolbutton.set_tooltip_text(tooltip)

        toolbutton.connect('clicked', callback)
        self.toolbar.insert(toolbutton, -1)
        toolbutton.show_all()

        return toolbutton

    def add_separator(self, position=None):
        """Adds a separator toolitem"""
        sep = SeparatorToolItem()
        if position is not None:
            self.toolbar.insert(sep, position)
        else:
            self.toolbar.insert(sep, -1)

        sep.show()

        return sep

    def remove(self, widget):
        """Removes a widget from the toolbar"""
        self.toolbar.remove(widget)

    # Callbacks (Uses the menubar's callback) #

    def on_toolbutton_add_clicked(self, data):
        log.debug('on_toolbutton_add_clicked')
        component.get('MenuBar').on_menuitem_addtorrent_activate(data)

    def on_toolbutton_remove_clicked(self, data):
        log.debug('on_toolbutton_remove_clicked')
        component.get('MenuBar').on_menuitem_remove_activate(data)

    def on_toolbutton_pause_clicked(self, data):
        log.debug('on_toolbutton_pause_clicked')
        component.get('MenuBar').on_menuitem_pause_activate(data)

    def on_toolbutton_resume_clicked(self, data):
        log.debug('on_toolbutton_resume_clicked')
        component.get('MenuBar').on_menuitem_resume_activate(data)

    def on_toolbutton_preferences_clicked(self, data):
        log.debug('on_toolbutton_preferences_clicked')
        component.get('MenuBar').on_menuitem_preferences_activate(data)

    def on_toolbutton_connectionmanager_clicked(self, data):
        log.debug('on_toolbutton_connectionmanager_clicked')
        component.get('MenuBar').on_menuitem_connectionmanager_activate(data)

    def on_toolbutton_queue_up_clicked(self, data):
        log.debug('on_toolbutton_queue_up_clicked')
        component.get('MenuBar').on_menuitem_queue_up_activate(data)

    def on_toolbutton_queue_down_clicked(self, data):
        log.debug('on_toolbutton_queue_down_clicked')
        component.get('MenuBar').on_menuitem_queue_down_activate(data)



#
# Copyright (C) 2009 Ian Martin <ianmartin@cantab.net>
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository import Gtk
from gi.repository.Gdk import RGBA

import deluge
from deluge import component
from deluge.common import fspeed
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client
from deluge.ui.gtk3.torrentdetails import Tab

from .common import get_resource
from .graph import Graph, size_formatter_scale

log = logging.getLogger(__name__)

# Gdk.RGBA textual spec
RED = 'rgb(255,0,0)'
GREEN = 'rgb(0,128,0)'
BLUE = 'rgb(0,0,255)'
DARKRED = 'rgb(139,0,0)'
ORANGE = 'rgb(255,165,0)'

DEFAULT_CONF = {
    'version': 2,
    'colors': {
        'bandwidth_graph': {'upload_rate': BLUE, 'download_rate': GREEN},
        'connections_graph': {
            'dht_nodes': ORANGE,
            'dht_cache_nodes': BLUE,
            'dht_torrents': GREEN,
            'num_connections': DARKRED,
        },
        'seeds_graph': {'num_peers': BLUE},
    },
}


def neat_time(column, cell, model, data):
    """Render seconds as seconds or minutes with label"""
    seconds = model.get_value(data, 0)
    if seconds > 60:
        text = '%d %s' % (seconds // 60, _('minutes'))
    elif seconds == 60:
        text = _('1 minute')
    elif seconds == 1:
        text = _('1 second')
    else:
        text = '%d %s' % (seconds, _('seconds'))
    cell.set_property('text', text)
    return


def int_str(number):
    return str(int(number))


def fspeed_shortform(value):
    return fspeed(value, shortform=True)


def text_to_rgba(color):
    """Turns a Color into a tuple with range 0-1 as used by the graph"""
    color_rgba = RGBA()
    color_rgba.parse(color)
    return color_rgba


class GraphsTab(Tab):
    def __init__(self, colors):
        super().__init__()

        builder = Gtk.Builder()
        builder.add_from_file(get_resource('tabs.ui'))
        self.window = builder.get_object('graph_tab')
        self.notebook = builder.get_object('graph_notebook')
        self.label = builder.get_object('graph_label')

        self._name = 'Stats'
        self._child_widget = self.window
        self._tab_label = self.label

        self.colors = colors

        self.bandwidth_graph = builder.get_object('bandwidth_graph')
        self.bandwidth_graph.connect('draw', self.on_graph_draw)

        self.connections_graph = builder.get_object('connections_graph')
        self.connections_graph.connect('draw', self.on_graph_draw)

        self.seeds_graph = builder.get_object('seeds_graph')
        self.seeds_graph.connect('draw', self.on_graph_draw)

        self.notebook.connect('switch-page', self._on_notebook_switch_page)

        self.selected_interval = 1  # Should come from config or similar
        self.select_bandwidth_graph()

        self.window.unparent()
        self.label.unparent()

        self.intervals = None
        self.intervals_combo = builder.get_object('combo_intervals')
        cell = Gtk.CellRendererText()
        self.intervals_combo.pack_start(cell, True)
        self.intervals_combo.set_cell_data_func(cell, neat_time)
        self.intervals_combo.connect('changed', self._on_selected_interval_changed)
        self.update_intervals()

    def on_graph_draw(self, widget, context):
        self.graph.draw_to_context(
            context,
            self.graph_widget.get_allocated_width(),
            self.graph_widget.get_allocated_height(),
        )
        # Do not propagate the event
        return True

    def update(self):
        d1 = client.stats.get_stats(list(self.graph.stat_info), self.selected_interval)
        d1.addCallback(self.graph.set_stats)

        def _update_complete(result):
            self.graph_widget.queue_draw()
            return result

        d1.addCallback(_update_complete)
        return d1

    def clear(self):
        pass

    def update_intervals(self):
        client.stats.get_intervals().addCallback(self._on_intervals_changed)

    def select_bandwidth_graph(self):
        log.debug('Selecting bandwidth graph')
        self.graph_widget = self.bandwidth_graph
        self.graph = Graph()
        colors = self.colors['bandwidth_graph']
        self.graph.add_stat(
            'download_rate',
            label='Download Rate',
            color=text_to_rgba(colors['download_rate']),
        )
        self.graph.add_stat(
            'upload_rate',
            label='Upload Rate',
            color=text_to_rgba(colors['upload_rate']),
        )
        self.graph.set_left_axis(
            formatter=fspeed_shortform, min=10240, formatter_scale=size_formatter_scale
        )

    def select_connections_graph(self):
        log.debug('Selecting connections graph')
        self.graph_widget = self.connections_graph
        g = Graph()
        self.graph = g
        colors = self.colors['connections_graph']
        g.add_stat('dht_nodes', color=text_to_rgba(colors['dht_nodes']))
        g.add_stat('dht_cache_nodes', color=text_to_rgba(colors['dht_cache_nodes']))
        g.add_stat('dht_torrents', color=text_to_rgba(colors['dht_torrents']))
        g.add_stat('num_connections', color=text_to_rgba(colors['num_connections']))
        g.set_left_axis(formatter=int_str, min=10)

    def select_seeds_graph(self):
        log.debug('Selecting connections graph')
        self.graph_widget = self.seeds_graph
        self.graph = Graph()
        colors = self.colors['seeds_graph']
        self.graph.add_stat('num_peers', color=text_to_rgba(colors['num_peers']))
        self.graph.set_left_axis(formatter=int_str, min=10)

    def set_colors(self, colors):
        self.colors = colors
        # Fake switch page to update the graph colors (HACKY)
        self._on_notebook_switch_page(
            self.notebook, None, self.notebook.get_current_page()  # This is unused
        )

    def _on_intervals_changed(self, intervals):
        liststore = Gtk.ListStore(int)
        for inter in intervals:
            liststore.append([inter])
        self.intervals_combo.set_model(liststore)
        try:
            current = intervals.index(self.selected_interval)
        except Exception:
            current = 0
        # should select the value saved in config
        self.intervals_combo.set_active(current)

    def _on_selected_interval_changed(self, combobox):
        model = combobox.get_model()
        tree_iter = combobox.get_active_iter()
        self.selected_interval = model.get_value(tree_iter, 0)
        self.update()
        return True

    def _on_notebook_switch_page(self, notebook, page, page_num):
        p = notebook.get_nth_page(page_num)
        if p is self.bandwidth_graph:
            self.select_bandwidth_graph()
            self.update()
        elif p is self.connections_graph:
            self.select_connections_graph()
            self.update()
        elif p is self.seeds_graph:
            self.select_seeds_graph()
            self.update()
        return True


class GtkUI(Gtk3PluginBase):
    def enable(self):
        log.debug('Stats plugin enable called')
        self.config = deluge.configmanager.ConfigManager(
            'stats.gtk3ui.conf', DEFAULT_CONF
        )

        self.builder = Gtk.Builder()
        self.builder.add_from_file(get_resource('config.ui'))

        component.get('Preferences').add_page(
            'Stats', self.builder.get_object('prefs_box')
        )
        component.get('PluginManager').register_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').register_hook(
            'on_show_prefs', self.on_show_prefs
        )
        self.on_show_prefs()

        self.graphs_tab = GraphsTab(self.config['colors'])
        self.torrent_details = component.get('TorrentDetails')
        self.torrent_details.add_tab(self.graphs_tab)

    def disable(self):
        component.get('Preferences').remove_page('Stats')
        component.get('PluginManager').deregister_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').deregister_hook(
            'on_show_prefs', self.on_show_prefs
        )
        self.torrent_details.remove_tab(self.graphs_tab.get_name())

    def on_apply_prefs(self):
        log.debug('applying prefs for Stats')
        gtkconf = {}
        for graph, colors in self.config['colors'].items():
            gtkconf[graph] = {}
            for value, color in colors.items():
                color_btn = self.builder.get_object(f'{graph}_{value}_color')
                try:
                    gtkconf[graph][value] = color_btn.get_color().to_string()
                except Exception:
                    gtkconf[graph][value] = DEFAULT_CONF['colors'][graph][value]
        self.config['colors'] = gtkconf
        self.graphs_tab.set_colors(self.config['colors'])

        config = {}
        client.stats.set_config(config)

    def on_show_prefs(self):
        for graph, colors in self.config['colors'].items():
            for value, color in colors.items():
                try:
                    color_btn = self.builder.get_object(f'{graph}_{value}_color')
                    color_btn.set_rgba(text_to_rgba(color))
                except Exception as ex:
                    log.debug('Unable to set %s %s %s: %s', graph, value, color, ex)
        client.stats.get_config().addCallback(self.cb_get_config)

    def cb_get_config(self, config):
        """Callback for on show_prefs."""
        pass



#
# Copyright (C) 2012 Bro <bro.development@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import base64

import pytest
import rencode

import deluge.log
from deluge.transfer import DelugeTransferProtocol

deluge.log.setup_logger('none')


class TransferTestClass(DelugeTransferProtocol):
    def __init__(self):
        DelugeTransferProtocol.__init__(self)
        self.transport = self
        self.messages_out = []
        self.messages_in = []
        self.packet_count = 0

    def write(self, message):
        """
        Called by DelugeTransferProtocol class
        This simulates the write method of the self.transport in DelugeTransferProtocol.
        """
        self.messages_out.append(message)

    def message_received(self, message):
        """
        This method overrides message_received is DelugeTransferProtocol and is
        called with the complete message as it was sent by DelugeRPCProtocol
        """
        self.messages_in.append(message)

    def get_messages_out_joined(self):
        return b''.join(self.messages_out)

    def get_messages_in(self):
        return self.messages_in

    def data_received_old_protocol(self, data):
        """
        This is the original method logic (as close as possible) for handling data receival on the client

        :param data: a zlib compressed string encoded with rencode.

        """
        import zlib

        print('\n=== New Data Received ===\nBytes received:', len(data))

        if self._buffer:
            # We have some data from the last dataReceived() so lets prepend it
            print('Current buffer:', len(self._buffer) if self._buffer else '0')
            data = self._buffer + data
            self._buffer = None

        self.packet_count += 1
        self._bytes_received += len(data)

        while data:
            print('\n-- Handle packet data --')

            print('Bytes received:', self._bytes_received)
            print('Current data:', len(data))

            if self._message_length == 0:
                # handle_new_message uses _buffer so set data to _buffer.
                self._buffer = data
                self._handle_new_message()
                data = self._buffer
                self._buffer = None
                self.packet_count = 1
                print('New message of length:', self._message_length)

            dobj = zlib.decompressobj()
            try:
                request = rencode.loads(dobj.decompress(data))
                print('Successfully loaded message', end=' ')
                print(
                    ' - Buffer length: %d, data length: %d, unused length: %d'
                    % (
                        len(data),
                        len(data) - len(dobj.unused_data),
                        len(dobj.unused_data),
                    )
                )
                print('Packet count:', self.packet_count)
            except Exception as ex:
                # log.debug('Received possible invalid message (%r): %s', data, e)
                # This could be cut-off data, so we'll save this in the buffer
                # and try to prepend it on the next dataReceived()
                self._buffer = data
                print(
                    'Failed to load buffer (size %d): %s' % (len(self._buffer), str(ex))
                )
                return
            else:
                data = dobj.unused_data
                self._message_length = 0

            self.message_received(request)


class TestDelugeTransferProtocol:
    @pytest.fixture(autouse=True)
    def set_up(self):
        """
        The expected messages corresponds to the test messages (msg1, msg2) after they've been processed
        by DelugeTransferProtocol.send, which means that they've first been encoded with rencode,
        and then compressed with zlib.
        The expected messages are encoded in base64 to easily including it here in the source.
        So before comparing the results with the expected messages, the expected messages must be decoded,
        or the result message be encoded in base64.

        """
        self.transfer = TransferTestClass()
        self.msg1 = (
            0,
            1,
            {'key_int': 1242429423},
            {'key_str': b'some string'},
            {'key_bool': True},
        )
        self.msg2 = (
            2,
            3,
            {'key_float': 12424.29423},
            {'key_unicode': 'some string'},
            {'key_dict_with_tuple': {'key_tuple': (1, 2, 3)}},
            {'keylist': [4, '5', 6.7]},
        )

        self.msg1_expected_compressed_base64 = (
            b'AQAAADF4nDvKwJjenp1aGZ+ZV+Lgxfv9PYRXXFLU'
            b'XZyfm6oAZGTmpad3gAST8vNznAEAJhSQ'
        )
        self.msg2_expected_compressed_base64 = (
            b'AQAAAF14nDvGxJzemZ1aGZ+Wk59Y4uTmpKib3g3il+ZlJuenpH'
            b'YX5+emKhSXFGXmpadPBkmkZCaXxJdnlmTEl5QW5KRCdIOZhxmB'
            b'hrUDuTmZxSWHWRpNnRyupaUBAHYlJxI='
        )

    def test_send_one_message(self):
        """
        Send one message and test that it has been sent correctoly to the
        method 'write' in self.transport.

        """
        self.transfer.transfer_message(self.msg1)
        # Get the data as sent by DelugeTransferProtocol
        messages = self.transfer.get_messages_out_joined()
        base64_encoded = base64.b64encode(messages)
        assert base64_encoded == self.msg1_expected_compressed_base64

    def test_receive_one_message(self):
        """
        Receive one message and test that it has been sent to the
        method 'message_received'.

        """
        self.transfer.dataReceived(
            base64.b64decode(self.msg1_expected_compressed_base64)
        )
        # Get the data as sent by DelugeTransferProtocol
        messages = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(messages)

    def test_receive_old_message(self):
        """
        Receive an old message (with no header) and verify that the data is discarded.

        """
        self.transfer.dataReceived(rencode.dumps(self.msg1))
        assert len(self.transfer.get_messages_in()) == 0
        assert self.transfer._message_length == 0
        assert len(self.transfer._buffer) == 0

    def test_receive_two_concatenated_messages(self):
        """
        This test simply concatenates two messsages (as they're sent over the network),
        and lets DelugeTransferProtocol receive the data as one string.

        """
        two_concatenated = base64.b64decode(
            self.msg1_expected_compressed_base64
        ) + base64.b64decode(self.msg2_expected_compressed_base64)
        self.transfer.dataReceived(two_concatenated)

        # Get the data as sent by DelugeTransferProtocol
        message1 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(message1)
        message2 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg2) == rencode.dumps(message2)

    def test_receive_three_messages_in_parts(self):
        """
        This test concatenates three messsages (as they're sent over the network),
        and lets DelugeTransferProtocol receive the data in multiple parts.

        """
        msg_bytes = (
            base64.b64decode(self.msg1_expected_compressed_base64)
            + base64.b64decode(self.msg2_expected_compressed_base64)
            + base64.b64decode(self.msg1_expected_compressed_base64)
        )
        packet_size = 40

        one_message_byte_count = len(
            base64.b64decode(self.msg1_expected_compressed_base64)
        )
        two_messages_byte_count = one_message_byte_count + len(
            base64.b64decode(self.msg2_expected_compressed_base64)
        )
        three_messages_byte_count = two_messages_byte_count + len(
            base64.b64decode(self.msg1_expected_compressed_base64)
        )

        for d in self.receive_parts_helper(msg_bytes, packet_size):
            bytes_received = self.transfer.get_bytes_recv()

            if bytes_received >= three_messages_byte_count:
                expected_msgs_received_count = 3
            elif bytes_received >= two_messages_byte_count:
                expected_msgs_received_count = 2
            elif bytes_received >= one_message_byte_count:
                expected_msgs_received_count = 1
            else:
                expected_msgs_received_count = 0
            # Verify that the expected number of complete messages has arrived
            assert expected_msgs_received_count == len(self.transfer.get_messages_in())

        # Get the data as received by DelugeTransferProtocol
        message1 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(message1)
        message2 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg2) == rencode.dumps(message2)
        message3 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(message3)

    # Remove underscore to enable test, or run the test directly:
    def _test_rencode_fail_protocol(self):
        """
        This test tries to test the protocol that relies on errors from rencode.

        """
        msg_bytes = (
            base64.b64decode(self.msg1_expected_compressed_base64)
            + base64.b64decode(self.msg2_expected_compressed_base64)
            + base64.b64decode(self.msg1_expected_compressed_base64)
        )
        packet_size = 149

        one_message_byte_count = len(
            base64.b64decode(self.msg1_expected_compressed_base64)
        )
        two_messages_byte_count = one_message_byte_count + len(
            base64.b64decode(self.msg2_expected_compressed_base64)
        )
        three_messages_byte_count = two_messages_byte_count + len(
            base64.b64decode(self.msg1_expected_compressed_base64)
        )

        print()

        print(
            'Msg1 size:',
            len(base64.b64decode(self.msg1_expected_compressed_base64)) - 4,
        )
        print(
            'Msg2 size:',
            len(base64.b64decode(self.msg2_expected_compressed_base64)) - 4,
        )
        print(
            'Msg3 size:',
            len(base64.b64decode(self.msg1_expected_compressed_base64)) - 4,
        )

        print('one_message_byte_count:', one_message_byte_count)
        print('two_messages_byte_count:', two_messages_byte_count)
        print('three_messages_byte_count:', three_messages_byte_count)

        for d in self.receive_parts_helper(
            msg_bytes, packet_size, self.transfer.data_received_old_protocol
        ):
            bytes_received = self.transfer.get_bytes_recv()

            if bytes_received >= three_messages_byte_count:
                expected_msgs_received_count = 3
            elif bytes_received >= two_messages_byte_count:
                expected_msgs_received_count = 2
            elif bytes_received >= one_message_byte_count:
                expected_msgs_received_count = 1
            else:
                expected_msgs_received_count = 0
            # Verify that the expected number of complete messages has arrived
            if expected_msgs_received_count != len(self.transfer.get_messages_in()):
                print(
                    'Expected number of messages received is %d, but %d have been received.'
                    % (
                        expected_msgs_received_count,
                        len(self.transfer.get_messages_in()),
                    )
                )

        # Get the data as received by DelugeTransferProtocol
        message1 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(message1)
        message2 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg2) == rencode.dumps(message2)
        message3 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(message3)

    def test_receive_middle_of_header(self):
        """
        This test concatenates two messsages (as they're sent over the network),
        and lets DelugeTransferProtocol receive the data in two parts.
        The first part contains the first message, plus two bytes of the next message.
        The next part contains the rest of the message.

        This is a special case, as DelugeTransferProtocol can't start parsing
        a message until it has at least 5 bytes (the size of the header) to be able
        to read and parse the size of the payload.

        """
        two_concatenated = base64.b64decode(
            self.msg1_expected_compressed_base64
        ) + base64.b64decode(self.msg2_expected_compressed_base64)
        first_len = len(base64.b64decode(self.msg1_expected_compressed_base64))

        # Now found the entire first message, and half the header of the next message  (2 bytes into the header)
        self.transfer.dataReceived(two_concatenated[: first_len + 2])

        # Should be 1 message in the list
        assert 1 == len(self.transfer.get_messages_in())

        # Send the rest
        self.transfer.dataReceived(two_concatenated[first_len + 2 :])

        # Should be 2 messages in the list
        assert 2 == len(self.transfer.get_messages_in())

        # Get the data as sent by DelugeTransferProtocol
        message1 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg1) == rencode.dumps(message1)
        message2 = self.transfer.get_messages_in().pop(0)
        assert rencode.dumps(self.msg2) == rencode.dumps(message2)

    # Needs file containing big data structure e.g. like thetorrent list as it is transfered by the daemon
    # def test_simulate_big_transfer(self):
    #    filename = '../deluge.torrentlist'
    #
    #    f = open(filename, 'r')
    #    data = f.read()
    #    message_to_send = eval(data)
    #    self.transfer.transfer_message(message_to_send)
    #
    # Get the data as sent to the network by DelugeTransferProtocol
    #    compressed_data = self.transfer.get_messages_out_joined()
    # packet_size = 16000 # Or something smaller...
    #
    #    for d in self.receive_parts_helper(compressed_data, packet_size):
    #        bytes_recv = self.transfer.get_bytes_recv()
    #        if bytes_recv < len(compressed_data):
    #            self.assertEqual(len(self.transfer.get_messages_in()), 0)
    #        else:
    #            self.assertEqual(len(self.transfer.get_messages_in()), 1)
    # Get the data as received by DelugeTransferProtocol
    #    transfered_message = self.transfer.get_messages_in().pop(0)
    # Test that the data structures are equal
    # self.assertEqual(transfered_message, message_to_send)
    # self.assertTrue(transfered_message == message_to_send)
    #
    # f.close()
    # f = open('rencode.torrentlist', 'w')
    # f.write(str(transfered_message))
    # f.close()

    def receive_parts_helper(self, data, packet_size, receive_func=None):
        byte_count = len(data)
        sent_bytes = 0
        while byte_count > 0:
            to_receive = packet_size if byte_count > packet_size else byte_count
            sent_bytes += to_receive
            byte_count -= to_receive
            if receive_func:
                receive_func(data[:to_receive])
            else:
                self.transfer.dataReceived(data[:to_receive])
            data = data[to_receive:]
            yield
